Unit ConfFiles;
{Conceptual unit that can proces a lot of different configfiles. (golded.cfg/
xconfig86 on one side, aliases, INI on the other)

   This is a unit of the XTDLIB/FPC package
   COPYRIGHT (C) 1993..1999 by Marco van de Voort,  All Rights Reserved
   Free for non-commercial use. See license.txt.
   (www.stack.nl/~marcov/xtdlib.htm)
   Mail for commercial info.

This is a raw version, meant to provoke feedback. The functionability is nice,
but it uses a lot of EPasStr routines, and is bulky.

A golded configurationfile is simply keywords with optional parameters, and
comments. The trick is that in these files, all the keywords are known.

An INI file you probably know, but this unit also supports subsections
([[subsection]] and [[[subsection]]]) INI file keys are not known by the
INI-file reader.

The other tested files (aliases and XFCONFIG86) are linux files in the /etc
directory. (XFCONFIG86 is X only)

The main routine of the below class simply reads the file, cuts out all
comments, and tries to find a keyword. If it knows the keyword, it looks
at the attributes in its definition (see KeywRec), and pre-processes the
optional argument(s). (itemisation, quotes etc). The preprocessed arguments
(the preprocessing can cut the argument line into pieces) are
passed with the Keywrec.CallRoutine value to the VIRTUAL method DoKeyword, which
you provide and should decide what to do with the data.
}

INTERFACE
{ $H-}
{$MODE OBJFPC}

USES EPasStr;   {XTDFPC unit which contains several string routines that are used}

CONST  MaxKeyW          =  80;   {Maximal number of keywords, can be increased}
       MaxKeywLength    =  31;   {Maximal length of a keyword}
       MaxItemSets      =  10;   {Maximal "SET OF CHAR"s used to automatically split
                                   a argument line into arguments}
       CommentIncr      = 512;   {If the comment PCHAR of a ConfigTree is too small,
                                   the allocation size is increased by CommentIncr
                                   bytes}
       MaxRecurDepth    =  10;   {Increase for idiotic complex file. Used as a safety
                                  measure to avoid useless deep recursion}
       ValueNotAvailable= 'N/O'; {Returned by property if not found.
                                   Length byte + 3 chars can be representated by a
                                   longint for easier testing is asm etc}

{I don't use enumerations for the series of constants below because descendants
might define additional ones, in case you wondered}

       {Predefined Item sets. Caution; if you use custom sets , these can be
       redefined}

       Item_Space             =0;
       Item_SpaceComma        =1;

       {Functions for preprocessing, if you have sane suggestions for other
         functions, mail me :-)}

       Prep_FuncNone          =0;
       Prep_GetPathParam      =1;            {Argument is single dirname or filename. Quotes supported}
       Prep_GetExistFileParam =2;            {Same as GetPathParam, and make sure it exists and
                                               is a file}
       Prep_GetExistDir       =3;            {Same as GetPathParam, and make sure it exists and
                                               is a Directory}
       Prep_GetNumberParam    =4;            {Get a plain LONGINT value}

       Prep_GetOctHexBin      =5;            {hex/oct/bin mode get value}

       {Some predefined errors, handled by the default ConfigError.
        Don't use values lower than 64, those are reserved for the main ConfFiles
        unit.
       }

       ConfErr_UnknownKeyword           =1;
       ConfErr_ZeroIsntAllowed          =2;
       ConfErr_TooManyKeywords          =3;
       ConfErr_Unbalanced_Quote         =4;
       ConfErr_Unexpected_Param         =5;
       ConfErr_File_Doesnt_Exist        =6;
       ConfErr_UndefinedFunction        =7;
       ConfErr_ParseErrorInConstant     =8;
       ConfErr_ErrorInSection           =9;
       ConfErr_ParamNotAlloc            =10;
       ConfErr_ErrInSectionNesting      =11;
       ConfErr_SyntaxError              =12;
       ConfErr_TooManyParam             =13;
       ConfErr_CustomErr                =14;


        { Type used to pass a read argument to the DoKeyword method}
TYPE ParamType   = RECORD
                    CASE ValueType   : LONGINT OF
                     0: (Value:String);
                     1: (Value2:LONGINT);
                    END;

     {Attributes that govern the way the argumentline of keywords are split up
     Upper      = Upcase() argumentline before further processing
     Lower      = LoCase() argumentline before further processing
     AllowZero  = Allow no arguments for this keyword
                  (otherwise error Zeroisntallowed is passed to ConfigError)
     Itemize    = Split up the argumentline using separators in a Set of Char
                           (=charset).
     QItemDouble= Split up the argumentline using separators in a Set of Char
                    (=charset). Separators within double quotes are ignored,
                    and the doublequoted string is considered one argument.
     QItemSingle= Split up the argumentline using separators in a Set of Char
                    (=charset). Separators within single quotes are ignored,
                    and the singlequoted string is considered one argument.
     QItemDoubleSingle= Same as QItemDouble and QItemSingle combined, allow
                   both single and double quoted arguments}

     FunctnsType=(upper,lower,AllowZero,Itemize,QItemDouble,QItemSingle,
                  QItemDoubleSingle);
     FunctnsSet = SET OF FunctnsType;

     {Definition of a keyword, and how to process it.
     Name              = Name of keyword
     CallRoutine       = nummeric value passed to DoKeyword to identify keyword
     Preprocess        = functions used to process the arguments when the
                          argumentline has been split up. (the prep_ constants above)
     Separator         = An index in an array of SET OF CHAR (=CHARSET) which contains
                          the separators for the Item functions of the Functns
                          field below
     Functns           = Functions used to preprocess the argumentline and split
                          it into the individual arguments}

     KeywRec=RECORD
              Name              : String[MaxKeywLength];
              CallRoutine       : LONGINT;
              Preprocess        : LONGINT;
              Separator         : LONGINT;
              Functns           : FunctnsSet;
              END;

      {TConfigTree structures. These structures describe the way TConfigTree
       stores the configurationfile. Descendants can operate directly on
       that "Tree", but using the properties is prefered}

      PParamData  = ^TParamDatRec;           {Single linked list that contains
                                              all arguments of a keyword}
      TParamDatRec= RECORD
                     NextParam   : PParamData;
                     Param       : ParamType;
                     END;

      PLineData   = ^TLineDataRec;           {Single linked list that contains
                                              all keywords and INI values in
                                              a section, plus all comments between
                                              them}
      TLineDataRec= RECORD
                     NextLineData: PLineData;
                     ParamData   : PParamData;  {Arguments of this keyword}
                     Comments    : PCHAR;       {Comments before this keyword}
                     KeyNo       : LONGINT;     {Keyword number or -1 which means
                                                 this is NOT a keyword but a inivalue}
                     ValueName   : String;      {Name of Ini value}
                     END;

      PSection    = ^TSectionRec;               { A section, brancing list}
      TSectionRec = RECORD
                     NextSection : PSection;    {Next section on this level}
                     SubSection  : PSection;    {Subsections of this section}
                     ParentSection : PSection;  {Link to parent section or NIL}
                     LineData    : PLineData;   {keywords/inivalues in this
                                                 section, but not those of this
                                                 sections subsections}
                     Comments    : PChar;       {Comments before the section command}
                     Level       : LONGINT;     {depth of this section. Root=1}
                     SectionName : String;      {Name of this section}
                     END;


{Class that reads a configuration file. It is complicated, but the main things
this class does is that
- all known keywords are passed to (virtual,abstract) method DoKeyword.
   If DoKeyword parameter "ItemNr"=0 then parameter "Params contain the first
   argument of a line, if itemnr=1 the second etc
- all [ini-sections] are passed to (virtual,abstract) method DoSection.
- all lines that are not comments, and don't contain ini-sections or keywords
  are passed to the (virtual,abstract) method Nonkeyword.
}

TYPE TConfigFile = CLASS
                      PRIVATE
                       FConf      : Text;               {File handle}
                       FInUse     : BOOLEAN;
                       ItemSetS   : ARRAY[0..MaxItemSets-1] OF CHARSET;

                       CommentBuffer: PCHAR;
                       CommentAllocated : LONGINT;

                     Protected
                      PROCEDURE HandleStrings(CONST InString:String;VAR keyword,arguments:String);

                       NrKeyW     : LONGINT;
                       KeyW       : ARRAY[1..MaxKeyW] OF KeywRec;

                       LongValue  : LONGINT;
                       ValueLoaded: BOOLEAN;
                       Line       : LONGINT;
                     PUBLIC
                      CONSTRUCTOR NewConfigFile; VIRTUAL;
                      DESTRUCTOR  KillConfigFile; VIRTUAL;
                       CommentSet : CHARSET;
                       KeySep     : CHARSET;

                       FUNCTION  ReadConfig(CONST Name:String) : LONGINT; VIRTUAL;      {Opens and processes config file}
                       PROCEDURE ReadBinValue(TextValue: String);
                       FUNCTION  CallRoutineToIndex(CallNumber:LONGINT):LONGINT;
                       PROCEDURE SetKeyWords(Keywords : ARRAY OF KeywRec);
                       PROCEDURE SetSeparators (ItemSetsIn : ARRAY OF CHARSET);
                       PROCEDURE CloseRFile;

                       {Virtual method with defaults}
                       PROCEDURE ConfigError(ErrNumber: LONGINT;Column:LONGINT;Data:String);VIRTUAL;

                       {Empty Virtual Methods, your code}
                       {Main method that processes a keyword}
                       FUNCTION DoKeyWord(KeywCode:LONGINT;Params:ParamType;ItemNr:LONGINT):LONGINT; VIRTUAL; ABSTRACT;
                       PROCEDURE DoSection(SectionName:String;Level:LONGINT); VIRTUAL; ABSTRACT;
                       PROCEDURE NonKeyword  (S:String); VIRTUAL; ABSTRACT;
                      END;

{TConfigTree hooks TConfigFiles DoKeyWord, DoSection and NonKeyword methods and
stores all information received from TConfigFiles (via these methods) in memory.
TConfigTree also provides methods to scan the data in memory.

Descendants may override the three methods above again and change the way how
the configurationfile is processed.
E.g. in demo READXF, sections are marked by known keyword "SECTION", which is
reported to DoKeyword. The descendant DoKeyword calls the TConfigTree DoKeyword
for all keywords except SECTION for which it calls DoSection, thus giving SECTION
the same functionality as the INI keyword.
}

      TConfigTree = CLASS (TConfigFile)
                     CONSTRUCTOR NewConfigFile; override;
                     DESTRUCTOR  KillConfigFile; override;
                     private
                      FUNCTION  DummyDoKeyWord(KeywCode:LONGINT;Params:ParamType;ItemNr:LONGINT): LONGINT;

                     protected
                      CurSection   : PSection;
                      CurLine      : PLineData;
                      CurParam     : PParamData;

                      {The property functions}
                      function  GlobalGet(KeyName:String;Number:LONGINT):String;
                      procedure GlobalPut(KeyName:String;Number:LONGINT;value:String);
                      FUNCTION  GlobalKeywordGet(Keyword:LONGINT;Number:LONGINT):String;
                      procedure GlobalKeywordPut(Keyword:LONGINT;Number:LONGINT;value:String);
                      FUNCTION  GlobalSectionGet(SectionName:String;Number:LONGINT):PSection;
                      function  LocalGet(KeyName:String;StartSection:PSection;Number:LONGINT):String;
                      procedure LocalPut(KeyName:String;StartSection:PSection;Number:LONGINT;value:String);
                      FUNCTION  LocalKeywordGet(Keyword:LONGINT;StartSection:PSection;Number:LONGINT):String;
                      procedure LocalKeywordPut(Keyword:LONGINT;StartSection:PSection;Number:LONGINT;value:String);
                      FUNCTION  LocalSectionGet(SectionName:String;StartSection:PSection;Number:LONGINT):PSection;
                      FUNCTION  LocalGenKeyGet(StartSection:PSection;Number:LONGINT):String;
                      PROCEDURE LocalGenKeyPut(StartSection:PSection;Number:LONGINT;value:String);
                      FUNCTION  GlobalGenKeyGet(Number:LONGINT):String;
                      PROCEDURE GlobalGenKeyPut(Number:LONGINT;value:String);

                      {These functions are the core of all properties}
                      FUNCTION  FindKeyword(KeyNumber:LONGINT;StartSection:PSection;Number:LONGINT;MaxLevel: LONGINT):String;
                      FUNCTION  FindKeyValue(Keyvalue:String;StartSection:PSection;Number:LONGINT;MaxLevel: LONGINT):String;
                      FUNCTION  FindSection(SectName:String;StartSection:Psection;Number:LONGINT;MaxLevel: LONGINT):PSection;
                     public
                      PROCEDURE KillParam(XParam:PParamData);
                      FUNCTION  ParamToString(XLine:PLineData):String;

                      property  GlobalKey[KeyName:String;Number:LONGINT]: string read GlobalGet write GlobalPut;
                      property  GlobalKeyword[Keyword:LONGINT;Number:LONGINT]:String read Globalkeywordget write Globalkeywordput;
                      property  GlobalSection[SectionName:String;Number:LONGINT]:psection read Globalsectionget;
                      property  GlobalGenericKey[Number:LONGINT]:string read GlobalGenKeyGet write GlobalGenKeyPut;
                      property  LocalKey[KeyName:String;StartSection:PSection;Number:LONGINT]: string read LocalGet write LocalPut;
                      property  LocalKeyword[Keyword:LONGINT;StartSection:PSection;Number:LONGINT]:String read Localkeywordget write Localkeywordput;
                      property  LocalSection[SectionName:String;StartSection:PSection;Number:LONGINT]:psection read Localsectionget;
                      property  LocalGenericKey[startsection:PSection;Number:LONGINT]:string read LocalGenKeyGet write localGenKeyPut;

                      {Function to get section nesting level in the possible inherited
                        versions of DoKeyWord..Nonkeyword, to avoid usage of CurSection if not
                        necessary}
                      FUNCTION GetLevel:LONGINT;

                      FUNCTION  DoKeyWord(KeywCode:LONGINT;Params:ParamType;ItemNr:LONGINT):LONGINT; VIRTUAL;override;
                      PROCEDURE DoSection(SectionName:String;Level:LONGINT); VIRTUAL; OVERRIDE;
                      PROCEDURE NonKeyword   (S:String); VIRTUAL; OVERRIDE;

                      {Only overrides to save trailing comments}
                      FUNCTION  ReadConfig(CONST Name:String) : LONGINT; OVERRIDE;

                      RootSection : PSection;
                      BackComment : PCHAR;        {Comments after last processed line in configfile}
                      IsIni      : BOOLEAN;        {If IsIni then tries to
                                                      separate non keyword lines into field=param}

                      PROCEDURE KillTree;
                      END;

{TConfigTreeView is the less generic one of the three classes. TConfigTreeView
is a standard partially configurable writer of the datastructures of
TConfigTree. It is good enough for not too complex formatted INI and/or keyword
based files.
}
         TConfigTreeView=CLASS(TConfigTree)
                          CONSTRUCTOR NewConfigFile; OVERRIDE;
                          Public
                            QuoteSet       : CHARSET;
                            FText          : Text;
                            KeywordIndent,
                            IniIndent,
                            ParamIndent    : LONGINT;
                            HoldSectionComment : BOOLEAN;
                            KeySepPref     : CHAR;
                           FUNCTION GetNextParam(VAR Handle:PParamData):ParamType;
                           PROCEDURE WriteComment(Comments : PChar);
                           PROCEDURE WriteValue(Handle:PLineData;Level:LONGINT); VIRTUAL;
                           PROCEDURE WriteSection(Section : PSection); VIRTUAL;
                           PROCEDURE BeforeExitFile; VIRTUAL;
                           PROCEDURE WriteTree(FileName:String);
                          END;

{Some generic functions}

FUNCTION FileExists(FileName: String): Boolean;
function IsDir(const S: String): Boolean;
function IsFile(const S: String): Boolean;

IMPLEMENTATION

Uses dos;

{--- Some generic proceures ---}

FUNCTION FileExists(FileName: String): Boolean;
{ Boolean function that returns True if the file exists;otherwise,
 it returns False. Closes the file if it exists, copied from the
 BP7 help, with a small change (filemode, avoids problems with some forms
 of sharing) }
var
 F: file;
 I : LONGINT;
begin
 {$I-}
 Assign(F, FileName);
 I:=FileMode;
 FileMode := 0;  {Set file access to read only, less problems with Win sharing}
 Reset(F);
 Close(F);
 {$I+}
 FileMode:=I;
 FileExists := (IOResult = 0) and (FileName <> '');
end;  { FileExists }


function IsDir(const S: String): Boolean;
var
  SR: SearchRec;
begin
  FindFirst(S, Directory, SR);
  if DosError = 0 then
    IsDir := SR.Attr and Directory <> 0
  else IsDir := False;
end;

function IsFile(const S: String): Boolean;
var
  SR: SearchRec;
begin
  FindFirst(S, Anyfile-directory, SR);
  if DosError = 0 then
    IsFile := SR.Attr and Directory <> 0
  else IsFile := False;
end;

function StringToParamType(S:String):ParamType;

VAR Param : ParamType;

BEGIN
 Param.ValueType:=0;
 Param.Value:=S;
 StringToParamType:=Param;
END;

{---------------- TConfigfile class ---------------------------------}

{$I configfi.inc}

{---------------- TConfigTree except properties (building the tree) -}

{$I conftree.inc}

{----------------- TConfigTree's Properties (Tree Scanning) ---------}

{$I conftrpr.inc}

{----------------- TConfigTreeView (writing the tree to file --------}

{$I conftrvw.inc}

END.