CONSTRUCTOR TConfigFile.NewConfigFile;

BEGIN
 NrKeyW:=0;
 FInUse:=FALSE;
 ItemSets[0]:=[' '];
 ItemSets[1]:=[',',' '];
 CommentSet:=[';'];
 KeySep:=['=',' '];
 GetMem(CommentBuffer,CommentIncr);
 CommentAllocated:=CommentIncr;
 ValueLoaded:=FALSE;
END;

DESTRUCTOR  TConfigFile.KillConfigFile;

BEGIN
 Freemem(CommentBuffer,CommentAllocated);
END;

PROCEDURE TConfigFile.CloseRFile;

BEGIN
 IF FInUse THEN
  Close(FConf);
 FInUse:=FALSE;
END;

FUNCTION  TConfigFile.CallRoutineToIndex(CallNumber:LONGINT):LONGINT;

VAR I : LONGINT;

BEGIN
 I:=1;
 WHILE (I<=NrKeyW) AND (Keyw[I].CallRoutine<>CallNumber) DO
  BEGIN
  INC(I);
  END;
 IF I>NrKeyW THEN       {Not Found}
  I:=0;
 CallRoutineTOIndex:=I;
END;

{Empty, since normal procedures do processing in the original class. Not
abstract since calling is allowed}

PROCEDURE TConfigFile.ReadBinValue(TextValue: String);
{Based on (own) Data2inc routines. Evaluate integers, max 32bit (hexadecimal,
octal,binary) constants).

Allowed styles :
Hexadecimal: $123 (Pascal style) 123H (M2 style) 123h 0x123 (C style)
Octal      : \xxx 2134o
Binary     : %1234
Decimal    : 123 123d 123D -123 -123d -123D
}

FUNCTION FixHex(CONST S1:String;base2:LONGINT):LONGINT;
{ Reads a base 2,8 or 16 constant from S1.
  Parameter = 2Log of base (1,3 or 4 corresponding to base 2,8 and 16)
  Constant is processed, the number of digits estimated (1,2 or 4 bytes) and
  the value is appended to msgtxt accordingly}

VAR I2,I3 : LONGINT;

BEGIN
 I3:=0;
 I2:=1;
 WHILE (S1[I2] IN ['0'..'9','A'..'F','a'..'f']) AND (I2<=Length(S1)) DO
  BEGIN
   IF (S1[I2]>#47) AND (S1[I2]<#58) THEN
    I3:=(I3 SHL base2)+ ORD(S1[I2])-48
   ELSE
    IF (S1[I2]>#64) AND (S1[I2]<#71) THEN
     I3:=(I3 SHL base2)+ ORD(S1[I2])-55
    ELSE
     IF (S1[I2]>#96) AND (S1[I2]<#103) THEN
      I3:=(I3 SHL base2)+ ORD(S1[I2])-87;
    INC(I2);
  END;
 FixHex:=I3;
END;

FUNCTION FixDec(CONST S1:String):LONGINT;
{ Reads decimal value starting at S1[1].

     Value in I3, number of digits found in I1}

VAR I1,I2,I3 : LONGINT;

BEGIN
 I1:=1;
 WHILE ((S1[I1]>#47) AND (S1[I1]<#58)) AND (I1<=Length(S1)) DO
  INC(I1);
 DEC(I1);
 IF I1=0 THEN
  ConfigError(ConfErr_ParseErrorInConstant,0,'');
 I3:=0;
 FOR I2:=1 TO I1 DO
  I3:=(I3*10)+ ORD(S1[I2])-48;
 FixDec:=I3;
END;

VAR Value : LONGINT;

BEGIN
 CASE TextValue[1] OF        {Select field type}
          '$' : BEGIN         {Handle $xxxx hex codes}
                 Delete(TextValue,1,1);
                 RTrim(TextValue,' ');
                 IF Length(TextValue)>0 THEN
                  Value:=FixHex(TextValue,4)
                 ELSE
                   ConfigError(ConfErr_ParseErrorInConstant,0,'');
                END;
         '-'  : BEGIN
                  Delete(TextValue,1,1);
                  Value:=-1*FixDec(TextValue);
                 END;
    '0'..'9'  : BEGIN    { handles 0x124,124124,124124H,234h,666o,353d,24b}
                 IF (Length(TextValue)>1) AND (TextValue[2]='x') THEN {C style 0xABCD hex}
                  BEGIN
                   Delete(TextValue,1,2);
                   Value:=FixHex(TextValue,4);
                  END
                 ELSE {other types (HP notation suffix h,o,d and b (and upcase versions,
                                                       and no suffix) }
                  BEGIN
                   CASE TextValue[Length(TextValue)] OF
                    'H','h' : Value:=FixHex(TextValue,4);      {Hex}
                    'o','O' : Value:=FixHex(TextValue,3);      {octal}
                    'B','b' : BEGIN       {Binary}
                               Delete(TextValue,Length(TextValue),1); {avoid 'b' char being treated as hex B }
                               Value:=FixHex(TextValue,1);
                              END;
           '0'..'9','d','D' : Value:=FixDec(TextValue);   {Fixdec is safe for trailing chars}
                             ELSE
                              ConfigError(ConfErr_ParseErrorInConstant,0,'');
                           END {Nested case}
                       END; { IF TextValue[2]='x'}
                      END; { '0'..'9'}
              '%'   : BEGIN          {%101010 binary constants}
                       Delete(TextValue,1,1);
                       Value:=FixHex(TextValue,1);
                      END;
              '\'   : BEGIN          {\xxx octal constants}
                       Delete(TextValue,1,1);
                       Value:=FixHex(TextValue,3);
                      END;
          END; {Case}
   ValueLoaded:=TRUE;
   LongValue:=Value;
END;

PROCEDURE TConfigFile.ConfigError(ErrNumber: LONGINT;Column:LONGINT;Data:String);

BEGIN
 IF FInUse THEN
  Close(FConf);
 CASE ErrNumber OF
  ConfErr_UnknownKeyword  : Writeln('Error: Unknown keyword ',Data,' in line ',Line,' .');
  ConfErr_ZeroIsntAllowed : Writeln('Error: Keyword in line ',Line,' requires an argument.');
  ConfErr_TooManyKeywords : Writeln('Error: The program tried to use more keywords than allowed');
  ConfErr_Unbalanced_Quote: Writeln('Error: Unbalanced quote in line ',Line);
  ConfErr_Unexpected_Param: Writeln('Error: Unexpected extra parameter "',Data,'" in line ',Line);
  ConfErr_File_Doesnt_Exist:Writeln('Error: File doesn'#39't exist in line ',Line);
  ConfErr_UndefinedFunction:Writeln('Error: Undefined function in line ',Line);
  ConfErr_ErrorInSection   :Writeln('Error: Invalid INI section near line ',Line);
  ConfErr_ParamNotAlloc    :Writeln('Error: Internal error in ConfigTree near line ',Line);
  ConfErr_ErrInSectionNesting: WriteLn('Error: Error in section nesting in line ',Line);
  ConfErr_SyntaxError      : WritelN('?Syntax Error in ',Line,'(',Data,')'#13#10'Ready');
  ConfErr_CustomErr        : Writeln(Data,' in line ',Line);
 ELSE
  Writeln('Error: Unknown error in configfile line ',Line);
  END;
 HALT;
END;

PROCEDURE TConfigFile.SetSeparators (ItemSetsin : ARRAY OF CHARSET);

VAR NrItems,I,J : LONGINT;

BEGIN
 NrItems:=HIGH(ItemSetsin )-Low(ItemSetsin )+1;
 IF NrItems> MaxItemSets THEN
  ConfigError(ConfErr_TooManyKeywords,0,'');
 IF NrItems> 0 THEN
  BEGIN
   I:=0;
   J:=LOW(ItemSetsin );
   WHILE J<=HIGH(ItemSetsin ) DO
    BEGIN
     ItemSets[I]:=ItemSetsin [J];

     INC(I);INC(J);
    END;
   END;
END;

PROCEDURE TConfigFile.HandleStrings(CONST InString:String;VAR keyword,arguments:String);

VAR A : LONGINT;

BEGIN
  A:=CharPosSet(InString,KeySep);                {Isolate keyword, INI compatible}
  IF A=0 THEN
   BEGIN
    Keyword:=UpCase(InString);                     {Only keyword}
    A:=Length(InString);
    Arguments:='';
   END
  ELSE
   BEGIN
    KeyWord  :=UpCase(Copy(InString,1,A-1));        {Space exists, cut keyword out}
    Arguments:=Copy(InString,A+1,Length(InString)-A);
    IF Arguments[1] IN CommentSet THEN                       {argument only space and comment?}
     Arguments:='';
   END;
  KillChar(Arguments,KeySep);
  RTrim(Arguments,' ');
//  LTrim(Arguments,' ');
END;

PROCEDURE TConfigFile.SetKeyWords(Keywords : ARRAY OF KeywRec);

VAR I,J : LONGINT;

BEGIN
 NrKeyW:=HIGH(Keywords)-Low(Keywords)+1;
 IF NrKeyW> MaxKeyW THEN
  ConfigError(ConfErr_TooManyKeywords,0,'');
 IF NrKeyW> 0 THEN
  BEGIN
   I:=1;
   J:=LOW(Keywords);
   WHILE J<=HIGH(Keywords) DO
    BEGIN
     KeyW[I]:=Keywords[J];
     KeyW[I].Name:=UPCASE(KeyW[I].Name);
     INC(I);INC(J);
    END;
   END;
END;

FUNCTION  TConfigFile.ReadConfig(CONST Name:String) : LONGINT;

PROCEDURE GetNummeric(CONST O : String);

VAR Code : LONGINT;

BEGIN
 VAL(O,Longvalue,Code);
 IF Code=0 THEN
  ValueLoaded:=TRUE
 ELSE
  ValueLoaded:=FALSE
END;

FUNCTION SearchKeyword(CONST Keyword:String):LONGINT;

VAR I: LONGINT;

BEGIN
 I:=0;
 IF NrKeyW>0 THEN
  BEGIN
   I:=1;
   REPEAT
    IF KeyW[I].Name=Keyword THEN
     Break;
    INC(I);
   UNTIL I>NrKeyW;
   IF I>NrKeyW THEN
    I:=0;
  END;
 SearchKeyword:=I;
END;

FUNCTION  GetFilNam(S:String):String;

VAR A : LONGINT;
    S2 : String;

BEGIN          {Get dirname or filename}
  IF S[1]='"' THEN
   BEGIN
    Delete(S,1,1);
    A:=Pos('"',S);
    IF A=0 THEN
     ConfigError(ConfErr_Unbalanced_Quote,0,'');
    S2:=Copy(S,1,A-1);
    Delete(S,1,A);
    LTrim(S,' ');
    IF (Length(S)>0) AND NOT (S[1] IN CommentSet) THEN
     ConfigError(ConfErr_Unexpected_Param,0,S);
    S:=S2;
   END
  ELSE
   BEGIN
    A:=CharPosSet(S,CommentSet);      {Not allowed outside quotes, so we can safely chop them off}
    IF A>0 THEN
     SetLength(S,Length(S)-1);
    A:=Pos(' ',S);
    IF A<>0 THEN
     ConfigError(ConfErr_Unexpected_Param,0,Copy(S,A,Length(S)-A));
   END;
   GetFilNam:=S;
END;

VAR KeyWNo : LONGINT;

PROCEDURE PreprocessParam(Itemedstring:String;Itemvalue:LONGINT);

VAR S : String;

BEGIN
 CASE Keyw[Keywno].Preprocess OF
         Prep_FuncNone            :  DoKeyword(Keyw[Keywno].CallRoutine,StringToParamType(ItemedString),itemvalue); {No preprocess}
         Prep_GetPathParam        : Begin
                                     S:=GetFilNam(ItemedString);
                                     DoKeyword(Keyw[Keywno].CallRoutine,StringToParamType(S),ItemValue);
                                    End;
         Prep_GetExistFileParam   : Begin
                                     S:=GetFilNam(ItemedString);
                                     IF NOT IsFile(S) THEN
                                      ConfigError(ConfErr_File_Doesnt_Exist,0,'');
                                     DoKeyword(Keyw[Keywno].CallRoutine,StringToParamType(S),itemvalue);
                                    End;
         Prep_GetExistDir       :   Begin {Get Dirname, and make sure it exists and is a dir}
                                     S:=GetFilNam(ItemedString);
                                     IF NOT IsDir(S) THEN
                                      ConfigError(ConfErr_File_Doesnt_Exist,0,'');
                                     DoKeyword(Keyw[Keywno].CallRoutine,StringToParamType(S),itemvalue);
                                    End;
         Prep_GetNumberParam    :   BEGIN
                                     GetNummeric(ItemedString);
                                     DoKeyword(Keyw[Keywno].CallRoutine,StringToParamType(ItemedString),itemvalue);
                                    END;
         Prep_GetOctHexBin      :   BEGIN
                                     ReadBinValue(ItemedString);
                                     DoKeyword(Keyw[Keywno].CallRoutine,StringToParamType(ItemedString),itemvalue);
                                    END;
        ELSE
         ConfigError(ConfErr_UndefinedFunction,0,'');
        END;
END;

Procedure DoQItem(QMode:LONGINT;CONST S : String);

VAR I,L : LONGINT;
    S2  : String;

BEGIN
 L:=0;
 REPEAT
  I:=QItem(S2,S,ItemSets[keyw[KeywNo].Separator],L,QMode);
  IF I=0 THEN
   BEGIN
    IF Length(S2)<>0 THEN
     PreprocessParam(S2,L);
   END
  ELSE
   ConfigError(ConfErr_Unbalanced_Quote,I-1,'');
  INC(L);
 UNTIL Length(S2)=0;
END;


VAR
    A,B    : LONGINT;
    S,Argument : String;
    Keyword    : String;
    P          : PChar;

BEGIN
 IF FileExists(Name) THEN
  BEGIN
   Assign(FConf,Name);
   Reset(FConf);
   FInUse:=TRUE;
   Line:=0;
   WHILE NOT EOF(FConf) DO
    BEGIN
     ReadLn(FConf,S);
     Inc(Line);
     LTrim(S,' ');
     IF (S[1] IN CommentSet) OR (Length(S)=0) THEN {Comment?}
      BEGIN
       SetLength(S,Length(S)+1);
       S[Length(S)]:=#10;                     {Append lf}
       A:=StrLen(CommentBuffer)+Length(S)+1;    {Calc needed space}
       IF CommentAllocated<=A THEN              {More needed}
        BEGIN
         B:= ((A DIV CommentIncr)+1)*CommentIncr;   {Calc next incr, resize}
         GetMem(P,B);
         Move (CommentBuffer[0],P[0],StrLen(CommentBuffer));
         Freemem(CommentBuffer,CommentAllocated);
         CommentBuffer:=P;
         CommentAllocated:=B;
        END;
       A:=StrLen(CommentBuffer);                {Append comments}
       Move(S[1],CommentBuffer[A],Length(S));
       CommentBuffer[A+Length(S)]:=#0;
      END
     ELSE
      BEGIN
       IF S[1]='[' THEN                                        {Ini-(Sub)section}
        BEGIN
         A:=1;
         WHILE (S[A]='[') AND (A<=Length(S)) DO
          INC(A);
         DEC(A);
         B:=Length(S);
         WHILE (S[B]=']') AND (B>0) DO
          DEC(B);
         B:=Length(S)-B;
         LTrim(S,'[');
         RTrim(S,']');
         IF A<>B THEN
          ConfigError(ConfErr_ErrorInSection,0,'');
         DoSection(S,A);                          { Try to create subsection level A}
        END
       ELSE
        BEGIN
         HandleStrings(S,Keyword,argument);
         KeyWNo:=SearchKeyword(Keyword);
         IF KeyWNo=0 THEN
           NonKeyword(S)
          ELSE
           BEGIN
            S:=Argument;
         IF (Length(S)=0) THEN
          BEGIN
           IF NOT (AllowZero IN KeyW[KeywNo].Functns) THEN
            ConfigError(ConfErr_ZeroIsntAllowed,0,'');
           DoKeyword(Keyw[Keywno].CallRoutine,StringToParamType(''),0);
          END
         ELSE
          BEGIN {Keyword known, commandline existing but 'raw'}
           IF (upper IN KeyW[KeywNo].Functns) THEN
            S:=UpCase(S);
           IF (lower IN KeyW[KeywNo].Functns) THEN
            Lowercase(S);
           IF Itemize IN KeyW[KeywNo].Functns THEN
            Begin {Get simple values separated by standard separate set}
             A:=0;
             REPEAT
              Item(Argument,S,itemsets[Keyw[keywno].separator],A);
              IF Length(Argument)<>0 THEN
               PreprocessParam(Argument,A);
              INC(A);
             UNTIL Length(Argument)=0;
           End
          ELSE
           IF QItemDouble IN KeyW[KeywNo].Functns THEN
            DoQItem(0,S)  {Get simple values sep'ed by standard separate set, which may
                                                  contain e.g. LFN's with separators within double quotes
                                                 e.g.  " lfn 1" "lfn 2" sfn3 "lfn 4"}
           ELSE
            IF QItemSingle IN KeyW[KeywNo].Functns THEN
             DoQItem(1,S)                           {Same as double, but for single quotes}
           ELSE
            IF QItemDoubleSingle IN KeyW[KeywNo].Functns THEN
             DoQItem(2,S)                           {Same as other two, both for
                                                   single and double quotes,
                                                   single within doubles etc etc}
          ELSE
           PreProcessParam(S,0);
          END; {Keyword known?}
         END;
        END;
       END;
  END; {WHILE}
 CloseRFile;
 ReadConfig:=1;
END
ELSE
 ReadConfig:=0;
END;
