{----------------- TConfigTree ----------------}

CONSTRUCTOR TConfigTree.NewConfigFile;

BEGIN
 Inherited NewConfigFile;
 NEW(RootSection);
 FillChar(RootSection^,SIZEOF(TSectionRec),#0);
 RootSection^.SectionName:='ROOTSECTION';
 RootSection^.Level:=1;
 CurSection:=RootSection;
END;

DESTRUCTOR  TConfigTree.KillConfigFile;

BEGIN
 KillTree;
 Inherited KillConfigFile;
END;

FUNCTION  TConfigTree.GetLevel : LONGINT;

BEGIN
 Result:=CurSection^.Level;
END;

FUNCTION  TConfigTree.ReadConfig(CONST Name:String) : LONGINT;

VAR I : LONGINT;

BEGIN
 I:=inherited ReadConfig(Name);
 IF (CommentBuffer<>NIL) AND (StrLen(CommentBuffer)<>0) THEN
  BEGIN
   GetMem(BackComment,StrLen(CommentBuffer)+1);
   Move(CommentBuffer[0],BackComment[0],StrLen(CommentBuffer)+1);
   CommentBuffer[0]:=#0;
  END;
 ReadConfig:=I;
END;

FUNCTION  TConfigTree.DummyDoKeyWord(KeywCode:LONGINT;Params:ParamType;ItemNr:LONGINT): LONGINT;

VAR S : String;
    A : LONGINT;

BEGIN
 IF ItemNr=0 THEN                    {First param, new Keyword/value}
  BEGIN
   IF CurSection^.LineData=NIL THEN
    BEGIN
     NEW(CurSection^.LineData);
     CurLine:=CurSection^.LineData;
    END
   ELSE
    BEGIN
     IF CurLine=NIL THEN
      BEGIN
       CurLine:=CurSection^.LineData;
       WHILE CurLine^.NextLineData<>NIL DO
        CurLine:=CurLine^.NextLineData;
      END;
     NEW(CurLine^.NextLineData);
     CurLine:=CurLine^.NextLineData;
    END;
   FillChar(CurLine^,SIZEOF(TLineDataRec),#0);
   IF StrLen(CommentBuffer)=0 THEN
    BEGIN
     CurLine^.Comments:=NIL;
    END
   ELSE
    BEGIN
     GetMem(CurLine^.CommentS,StrLen(CommentBuffer)+1);
     Move(CommentBuffer[0],CurLine^.CommentS[0],StrLen(CommentBuffer)+1);
     CommentBuffer[0]:=#0;
    END;
   IF (keywcode<>-1) OR IsIni THEN
    BEGIN
     NEW(CurLine^.ParamData);
     CurParam:=CurLine^.ParamData;
     FillChar(CurParam^,SIZEOF(TParamDatRec),#0);
    END
   ELSE
    CurParam:=NIL;
   IF KeywCode<>-1 THEN                                 {Standard keyword}
    BEGIN
     CurLine^.KeyNo:=Keywcode;
     CurLine^.ValueName:='';
     CurParam^.Param:=Params;
    END
   ELSE
    BEGIN                                               {Non-standard keyword}
     CurLine^.KeyNo:=-1;
     IF IsIni THEN                                      {Try to split like INI?}
      BEGIN
       A:=CharPosSet(Params.Value,KeySep);
       IF A<>0 THEN                                     {split OK!}
        BEGIN
         CurLine^.ValueName:=Copy(Params.Value,1,A-1);
         RTrim(CurLine^.ValueName,' ');
         Delete(Params.Value,1,A);
         KillChar(Params.Value,[' ',#9]);
         IF Params.Value[1]='"' THEN
          BEGIN
           Delete(Params.Value,1,1);
           IF NOT (Params.Value[Length(Params.Value)]='"') THEN
            ConfigError(ConfErr_Unbalanced_Quote,0,'');
           Delete(Params.Value,Length(Params.Value),1);
          END;
         CurParam^.Param:=Params;
        END
       ELSE
         CurLine^.ValueName:=Params.Value;              {Params stays empty}
     END {IsIni}
    ELSE
     CurLine^.ValueName:=Params.Value;              {Params not allocated}
   END; {Keywcode<>-1}
  END
 ELSE { Itemnr<>0}
  BEGIN
   IF (CurParam=NIL) OR (CurLine=NIL) THEN
    ConfigError(ConfErr_ParamNotAlloc ,0,'');
   NEW(CurParam^.NextParam);
   CurParam:=CurParam^.NextParam;
   CurParam^.NextParam:=NIL;
   CurParam^.Param:=Params;
  END;
END;

FUNCTION  TConfigTree.DoKeyWord(KeywCode:LONGINT;Params:ParamType;ItemNr:LONGINT): LONGINT;

BEGIN
 DoKeyWord:=DummyDoKeyWord(KeywCode,Params,ItemNr);
END;

PROCEDURE TConfigTree.DoSection(SectionName:String;Level:LONGINT);

VAR RetVal : LONGINT;
    Old    : PSection;

PROCEDURE NewSectionAtThisLevel;

BEGIN
  NEW(CurSection^.NextSection);
  Old:=CurSection;
  CurSection:=CurSection^.NextSection;
  FillChar(CurSection^,SIZEOF(TSectionRec),#0);
  CurSection^.ParentSection:=Old^.ParentSection;
  CurSection^.Level:=Old^.Level;
END;

PROCEDURE FixCommentAndName;

VAR  A : LONGINT;

BEGIN
  CurSection^.SectionName:=SectionName;
  A:=StrLen(CommentBuffer)+1;
  IF A<>1 THEN
   BEGIN
    GetMem(CurSection^.Comments,A);
    Move(CommentBuffer[0],CurSection^.Comments[0],A);
    CommentBuffer[0]:=#0;
   END
END;

BEGIN
   RetVal:=Level-CurSection^.Level;
   CASE RetVal OF
    0  :  BEGIN                             {New Level on same niveau}
           NewSectionAtThisLevel;
           FixCommentAndName;
           END;
    1 :   BEGIN
           Old:=CurSection;
           IF CurSection^.SubSection=NIL THEN
            BEGIN
             NEW(CurSection^.SubSection);
             CurSection:=CurSection^.SubSection;
            END
           ELSE
            BEGIN
             CurSection:=CurSection^.Subsection;
             WHILE CurSection^.NextSection<>NIL DO
              CurSection:=CurSection^.NextSection;
             NEW(CurSection^.NextSection);
             CurSection:=CurSection^.Nextsection;
            END;
           FillChar(CurSection^,SIZEOF(TSectionRec),#0);
           CurSection^.ParentSection:=Old;
           CurSection^.Level:=Old^.Level+1;
           FixCommentAndName;
         END;
    ELSE
     BEGIN
       IF RetVal>1 THEN
        ConfigError(ConfErr_ErrInSectionNesting,0,'');  {Can't skip levels}
       RetVal:=-RetVal;
       WHILE RetVal>0 DO
        BEGIN
         IF CurSection^.ParentSection=NIL THEN
          ConfigError(ConfErr_ErrInSectionNesting,0,'');
         CurSection:=CurSection^.ParentSection;
         DEC(RetVal);
        END;
       IF Length(SectionName)=0 THEN
        BEGIN
         IF CurSection^.LineData=NIL THEN
           CurLine:=NIL
         ELSE
          BEGIN
          CurLine:=CurSection^.LineData;
          WHILE CurLine^.NextLineData<>NIL DO
           CurLine:=CurLine^.NextLineData;
          END;
        END
       ELSE
        BEGIN
         NewSectionAtThisLevel;
         FixCommentAndName;
        END;
      END; {ELSE Case}
     END;
END;

PROCEDURE TConfigTree.NonKeyword   (S:String);

VAR VarRet : LONGINT;

BEGIN {!!}
{   {TConfigTree.}DoKeyword(-1,StringToParamType(S),0);   {Calling just
                                                        DoKeyword can introduce loops}}
  DummyDoKeyWord(-1,StringToParamType(S),0);
END;

FUNCTION TConfigTree.ParamToString(XLine:PLineData):String;

VAR S: String;
    P : PParamData;

BEGIN
 IF (XLine<>NIL) AND (XLine^.ParamData<>NIL) THEN
  BEGIN
   P:=XLine^.ParamData;
   S:='';
   WHILE P<>NIL DO
    BEGIN
     S:=S+P^.Param.Value+' ';
     P:=P^.NextParam;
    END;
  END
 ELSE
  S:='';
 RTrim(S,' ');
 ParamToString:=S;
END;

PROCEDURE TConfigTree.KillParam(XParam:PParamData);

BEGIN
 IF XParam<>NIL THEN
  BEGIN
   if Xparam^.NextParam<>NIL THEN
    KillParam(XParam^.Nextparam);
  {$IFDEF ShowFreememBug}
   FreeMem(XParam,SIZEOF(TParamDatRec));
  {$ELSE}
  DISPOSE(XParam);
  {$ENDIF}
 END;
END;

PROCEDURE TConfigTree.KillTree;

PROCEDURE KillLine(XLine:PLineData);

BEGIN
 IF XLine^.Comments<>NIL THEN
  FreeMem(XLine^.Comments,StrLen(XLine^.Comments)+1);
 IF XLIne^.ParamData<>NIL THEN
  KillParam(XLine^.ParamData);
 IF XLine^.NextLineData<>NIL THEN
  KillLine(XLine^.NextLineData);
 {$IFDEF ShowFreememBug}
  FreeMem(XLine,SIZEOF(TLineDataRec));
 {$ELSE}
  Dispose(XLine);
 {$ENDIF}

END;

PROCEDURE KillSection(Section:PSection);

BEGIN
 IF Section^.Comments<>NIL THEN
  FreeMem(Section^.Comments,StrLen(Section^.Comments)+1);
 IF Section^.LineData<>NIL THEN
  KillLine(Section^.LineData);
 IF Section^.SubSection<>NIL THEN
  KillSection(Section^.SubSection);
 IF Section^.NextSection<>NIL THEN
  KillSection(Section^.NextSection);
 {$IFDEF ShowFreememBug}
  FreeMem(Section,SIZEOF(TSectionRec)); {**}
 {$ELSE}
  DISPOSE(Section);
 {$ENDIF}

END;

BEGIN
 IF RootSection<>NIL THEN
  BEGIN
   KillSection(RootSection);
   IF BackComment<>NIL THEN
    FreeMem(BackComment,StrLen(BackComment)+1);
   RootSection:=NIL;
  END;
END;
