{-------------- TConfigTree Properties (Tree Scanning) ---------------}

CONST MaxSectionDepth=20;       {The functions don't recurse deeper than this}

FUNCTION TConfigTree.FindKeyValue(Keyvalue:String;StartSection:PSection;Number:LONGINT;MaxLevel: LONGINT):String;
{Finds a key in section StartSection and below. The procedure recursively searches maximally "Maxlevel"
levels beyond the startsection level}

VAR SaveString : String;
    IKeyw,IArgS : String;

FUNCTION SearchLineData(XLine:PLineData):LONGINT;

VAR  A,B : LONGINT;
     S   : String;

BEGIN
 A:=0;
 IF XLine <>NIL THEN
  BEGIN
  IF XLine^.KeyNo=-1 THEN       {Not a keyword but a ini'sh key=value structure}
   BEGIN
    HandleStrings(XLine^.ValueName,IKeyw,IArgS);
    Writeln(XLine^.ValueName);
    IF IKeyw=KeyValue THEN
     BEGIN
      DEC(Number);
      Writeln(Number);
      IF Number=0 THEN
       BEGIN
        SaveString:=IArgS+' '+ParamToString(XLine);
        CurLine:=XLine;
        EXIT(1);
       END;
     END;
   END;
  IF XLine^.NextLineData<>NIL THEN
   A:=SearchLineData(XLine^.NextLineData);
 END;
 SearchLineData:=A;
END;

FUNCTION FindLine_Section(Section:PSection):LONGINT;

VAR Found: LONGINT;

BEGIN
 IF SearchLineData(Section^.LineData)<>1 THEN
  BEGIN {Not found in values of this section. Continue to search other sections}
   Found:=0;
   IF (Section^.Level<=MaxLevel) AND (Section^.SubSection<>NIL) THEN
    Found:=FindLine_Section(Section^.SubSection);
   IF (Found<>0) AND (Section^.NextSection<>NIL) THEN
    Found:=FindLine_Section(Section^.NextSection);
   FindLine_Section:=Found;
  END
 ELSE
  BEGIN
   Found:=1;
   CurSection:=Section;
  END;
 FindLine_Section:=Found;
END;

VAR Found : LONGINT;

BEGIN
 CurLine:=NIL;
 CurSection:=NIL;
 IF StartSection<>NIL THEN
  BEGIN
   SaveString:='';
   INC(MaxLevel,StartSection^.Level-1);         {Relative to absolute level}
   KeyValue:=UPCASE(KeyValue);
   IF FindLine_Section(StartSection)=1 THEN
    FindKeyValue:=SaveString
   ELSE
    FindKeyValue:=ValueNotAvailable;
  END
  ELSE
   FindKeyValue:=ValueNotAvailable;
END;

FUNCTION TConfigTree.FindKeyword(KeyNumber:LONGINT;StartSection:PSection;Number:LONGINT;MaxLevel: LONGINT):String;
{Finds a "number"th keyword "KeyNumber" in section StartSection and below. The procedure recursively searches maximally "Maxlevel"
levels beyond the startsection level}

{VAR FoundLine   : PLineData;
    FoundSection: PSection;}

FUNCTION SearchLineData(XLine:PLineData):LONGINT;

VAR  A : LONGINT;

BEGIN
 A:=0;
 IF XLine <>NIL THEN
  BEGIN
  IF XLine^.KeyNo=KeyNumber THEN
   BEGIN
    DEC(Number);
    IF Number=0 THEN
     BEGIN
      CurLine:=XLine;
      EXIT(1);
     END;
   END;
  IF XLine^.NextLineData<>NIL THEN
   A:=SearchLineData(XLine^.NextLineData);
 END;
 SearchLineData:=A;
END;

FUNCTION FindLine_Section(Section:PSection):LONGINT;

VAR Found: LONGINT;

BEGIN
 IF SearchLineData(Section^.LineData)<>1 THEN
  BEGIN {Not found in values of this section. Continue to search other sections}
   Found:=0;
   IF (Section^.Level<=MaxLevel) AND (Section^.SubSection<>NIL) THEN
    Found:=FindLine_Section(Section^.SubSection);
   IF (Found=0) AND (Section^.NextSection<>NIL) THEN
     Found:=FindLine_Section(Section^.NextSection);
   FindLine_Section:=Found;
  END
 ELSE
  BEGIN
   Found:=1;
   CurSection:=Section;
  END;
 FindLine_Section:=Found;
END;

VAR Found : LONGINT;

BEGIN
 CurLine:=NIL;
 CurSection:=NIL;
 IF StartSection<>NIL THEN
  BEGIN
   INC(MaxLevel,StartSection^.Level-1);         {Convert relative to absolute level}
   IF FindLine_Section(StartSection)=1 THEN
     FindKeyword:=ParamToString(CurLine)
   ELSE
    FindKeyword:=ValueNotAvailable;
  END
  ELSE
   FindKeyword:=ValueNotAvailable;
END;

FUNCTION TConfigTree.FindSection(SectName:String;StartSection:Psection;Number:LONGINT;MaxLevel: LONGINT):PSection;

FUNCTION FindSect_Section(Section:PSection):PSection;

VAR Found: PSection;

BEGIN
 IF UpCase(Section^.SectionName)=SectName THEN
  BEGIN
   DEC(Number);
   IF Number=0 THEN
    BEGIN
     FindSect_Section:=Section;
     EXIT;
    END;
   END;
 Found:=NIL;
 IF (Section^.Level<=MaxLevel) AND (Section^.SubSection<>NIL) THEN
   Found:=FindSect_Section(Section^.SubSection);
 IF (Found=NIL) AND (Section^.NextSection<>NIL) THEN
  Found:=FindSect_Section(Section^.NextSection);
 FindSect_Section:=Found;
END;

BEGIN
 SectName:=UPCASE(SectName);
 IF StartSection<>NIL THEN
  FindSection:=FindSect_Section(StartSection)
 ELSE
  FindSection:=NIL;
END;

{--- Methods for the GlobalKey ---}

FUNCTION TConfigTree.GlobalGet(KeyName:String;Number:LONGINT):String;

BEGIN
 GlobalGet:=FindKeyValue(KeyName,RootSection,Number,MaxSectionDepth)
END;

PROCEDURE TConfigTree.GlobalPut(KeyName:String;Number:LONGINT;value:String);

BEGIN
 FindKeyValue(KeyName,RootSection,Number,MaxSectionDepth);
 {If Found, CurLine is loaded with pointer to the Numberth occurence}
 IF CurLine<>NIL THEN
  BEGIN
   IF CurLine^.ParamData=NIL THEN
     NEW(CurLine^.ParamData)
   ELSE
     KillParam(CurLine^.ParamData^.NextParam);
   FillChar(CurLine^.ParamData,SIZEOF(TParamDatRec),#0);
   CurLine^.ParamData^.Param.Value:=value;
  END;
END;

{--- Methods for the GlobalKeyword ---}

FUNCTION TConfigTree.GlobalKeywordGet(Keyword:LONGINT;Number:LONGINT):String;

BEGIN
 GlobalKeywordGet:=FindKeyword(Keyword,RootSection,Number,MaxSectionDepth)
END;

PROCEDURE TConfigTree.GlobalKeywordPut(Keyword,Number:LONGINT;value:String);

BEGIN
 FindKeyword(Keyword,RootSection,Number,MaxSectionDepth);
 {If Found, CurLine is loaded with pointer to the Numberth occurence}
 IF CurLine<>NIL THEN
  BEGIN
   IF CurLine^.ParamData=NIL THEN
     NEW(CurLine^.ParamData)
   ELSE
     KillParam(CurLine^.ParamData^.NextParam);
   FillChar(CurLine^.ParamData,SIZEOF(TParamDatRec),#0);
   CurLine^.ParamData^.Param.Value:=value;
  END;
END;

{--- Methods for the GlobalSection ---}

FUNCTION TConfigTree.GlobalSectionGet(SectionName:String;Number:LONGINT):PSection;

BEGIN
 GlobalSectionGet:=FindSection(SectionName,RootSection,Number,MaxSectionDepth)
END;

{--- Methods for the LocalKey ---}

FUNCTION TConfigTree.LocalGet(KeyName:String;StartSection:PSection;Number:LONGINT):String;

BEGIN
 LocalGet:=FindKeyValue(KeyName,StartSection,Number,MaxSectionDepth)
END;

PROCEDURE TConfigTree.LocalPut(KeyName:String;StartSection:PSection;Number:LONGINT;value:String);

BEGIN
 FindKeyValue(KeyName,StartSection,Number,MaxSectionDepth);
 {If Found, CurLine is loaded with pointer to the Numberth occurence}
 IF CurLine<>NIL THEN
  BEGIN
   IF CurLine^.ParamData=NIL THEN
     NEW(CurLine^.ParamData)
   ELSE
     KillParam(CurLine^.ParamData^.NextParam);
   FillChar(CurLine^.ParamData,SIZEOF(TParamDatRec),#0);
   CurLine^.ParamData^.Param.Value:=value;
  END;
END;

{--- Methods for the LocalKeyword ---}

FUNCTION TConfigTree.LocalKeywordGet(Keyword:LONGINT;StartSection:PSection;Number:LONGINT):String;

BEGIN
 LocalKeywordGet:=FindKeyword(Keyword,StartSection,Number,MaxSectionDepth)
END;

PROCEDURE TConfigTree.LocalKeywordPut(Keyword:LONGINT;StartSection:PSection;Number:LONGINT;value:String);

BEGIN
 FindKeyword(Keyword,StartSection,Number,MaxSectionDepth);
 {If Found, CurLine is loaded with pointer to the Numberth occurence}
 IF CurLine<>NIL THEN
  BEGIN
   IF CurLine^.ParamData=NIL THEN
     NEW(CurLine^.ParamData)
   ELSE
     KillParam(CurLine^.ParamData^.NextParam);
   FillChar(CurLine^.ParamData,SIZEOF(TParamDatRec),#0);
   CurLine^.ParamData^.Param.Value:=value;
  END;
END;

{--- Methods for the LocalSection ---}

FUNCTION TConfigTree.LocalSectionGet(SectionName:String;StartSection:PSection;Number:LONGINT):PSection;

BEGIN
 LocalSectionGet:=FindSection(SectionName,StartSection,Number,MaxSectionDepth)
END;

{---- Methods for accessing the keys in the binary tree as one array ----}

FUNCTION TConfigTree.LocalGenKeyGet(StartSection:PSection;Number:LONGINT):String;

BEGIN
 LocalGenKeyGet:=FindKeyword(-1,StartSection,Number,MaxSectionDepth);
END;

PROCEDURE TConfigTree.LocalGenKeyPut(StartSection:PSection;Number:LONGINT;value:String);

BEGIN
 FindKeyword(-1,StartSection,Number,MaxSectionDepth);
 {If Found, CurLine is loaded with pointer to the Numberth occurence}
 IF CurLine<>NIL THEN
  BEGIN
   IF CurLine^.ParamData=NIL THEN
     NEW(CurLine^.ParamData)
   ELSE
     KillParam(CurLine^.ParamData^.NextParam);
   FillChar(CurLine^.ParamData,SIZEOF(TParamDatRec),#0);
   CurLine^.ParamData^.Param.Value:=value;
  END;
END;

FUNCTION TConfigTree.GlobalGenKeyGet(Number:LONGINT):String;

BEGIN
 FindKeyword(-1,RootSection,Number,MaxSectionDepth);
 IF CurLine<>NIL THEN
  GlobalGenKeyGet:=CurLine^.ValueName
 ELSE
  GlobalGenKeyGet:=ValueNotAvailable;
END;

PROCEDURE TConfigTree.GlobalGenKeyPut(Number:LONGINT;value:String);

BEGIN
 FindKeyword(-1,RootSection,Number,MaxSectionDepth);
 {If Found, CurLine is loaded with pointer to the Numberth occurence}
 IF CurLine<>NIL THEN
  BEGIN
   IF CurLine^.ParamData=NIL THEN
     NEW(CurLine^.ParamData)
   ELSE
     KillParam(CurLine^.ParamData^.NextParam);
   FillChar(CurLine^.ParamData,SIZEOF(TParamDatRec),#0);
   CurLine^.ParamData^.Param.Value:=value;
  END;
END;
