Program CookCfg;
{Example program for ConfFiles unit. In the future this will be the core of the
 configuration reader for cookie. Some routines (like DoBlock and subprocs) will
 be shared though, because the cookie project has several configuration files.


   This is a program of the XTDLIB/FPC package
   COPYRIGHT (C) 1993..1998 by Marco van de Voort,  All Rights Reserved
   Free for non-commercial use. See license.txt.
   (www.stack.nl/~marcov/xtdlib.htm) 
   Mail for commercial info.

This is (until now) the only example with a configurationfile designed more
or less for use with Conffiles. Actually, the basic TConfigFile class is a
redesigned, generalized version of code which was once used in this demo.

Also this is the only example that uses TConfigFile as a base to build a
system, that only reads a configurationfile.

The principle difference between this demo, and the other ones is that
1 CookCfg entirely processes the configfile while converting to structures
  native to the application, while
2 the other ones (readxf,initest,aliasreader) read the configfile in a
   datastructure native to the ConfFiles unit, which is quered by the
   application by the ConfigTree properties, and properties on top of those.

Method 1 above, is much faster than 2, but 2 allows (quite easy) modifications,
and writing back to file. Method 1 requires an additional writer, which never
can generate the exact file back.

I chose method 1 for cookie because some of the datastructures are parsed
possibly a 100000 times per run. That is also the reason for the ugly
construct for the categories and subcats fields of ConfigRecord. The use
of two dynamically allocated arrays instead of an array of record or linked list
allows to use scasw for lookups in the category array, if the processing speed
turns out to be too low.
}

Uses ConfFiles,EPasstr; {cooktypes,Cookglobals}

TYPE CookieKeywEnum=(Key_Block,Key_DefCookie, key_DefOffensive,
                     key_Alldisabled, Key_AllEnabled,key_allow);



CONST
    MaxConfig=10;
    GrowGap  =32;
    CookErr_ErrInBlockItem       =64;
    CookErr_MoreConfigThanAllowed=65;




ExampleKeywords : ARRAY[key_block..Key_allow] OF KeywRec=(
(name:'block';               CallRoutine:ORD(Key_Block);           Preprocess:prep_GetOctHexBin; Separator : Item_spacecomma;
                             Functns:[Itemize]),
(name:'defaultcookiefile';   CallRoutine:ORD(Key_DefCookie);       Preprocess:Prep_GetPathParam; Separator : Item_space;
                             Functns:[]),
(name:'defaultoffensivefiles';CallRoutine:ORD(Key_DefOffensive);   Preprocess:Prep_GetPathParam; Separator : Item_spacecomma;
                             Functns:[QItemDouble]),
(name:'alldisabled';         CallRoutine:ORD(Key_AllDisabled);     Preprocess:0;                 Separator : Item_space;
                             Functns:[AllowZero]),
(name:'allenabled';          CallRoutine:ORD(Key_AllEnabled);      Preprocess:0;                 Separator : Item_space;
                             Functns:[AllowZero]),
(name:'Allow';               CallRoutine:ORD(Key_Allow);           Preprocess:0;                 Separator : Item_spacecomma;
                             Functns:[itemize]));


TYPE  PKeyRecord = ^Keyrecord;
      KeyRecord=RECORD
                 Next  : PkeyRecord;
                 Name  : String[30];
                 Value : String;
                 END;
      PCategSelect   =^CategSelectRec;
      CategSelectRec = RECORD
                        Category : LONGINT;
                        Lower,     { Lower> Upper means 0..upper lower..32767}
                        Upper    : LONGINT;
                        END;
      PConfigRecord= ^ConfigRecord;
      ConfigRecord = RECORD
                      ConfigName    : String[20];
                      DefCookie,
                      DefOffensive  : String;
                      AllEnabled    : BOOLEAN;  {Not all enabled= all disabled}
                      EnableTriggered:BOOLEAN;
//                    OtherKeyroot  : PKeyRecord;
                      Categories    : ^LONGINT;
                      SubCats       : ^LONGINT;
                      CatAllocated  : LONGINT; { elements allocated for Categories and SubCats}
                      CatUsed       : LONGINT; { elements actually used}


                      BigSet        : ARRAY[0..2] OF SET OF BYTE;
                      END;

     TCookieConfig=CLASS(TConfigFile)
                    private
                     NrConfigs  : LONGINT;
                     Configs    : ARRAY[0..MaxConfig-1] OF PConfigRecord;
                     CurKeys    : PKeyRecord;
                     CurCateg   : PCategSelect;
                    public
                     CONSTRUCTOR NewConfigFile; OVERRIDE;
                     PROCEDURE   ConfigError(ErrNumber: LONGINT;Column:LONGINT;Data:String); OVERRIDE;
                     FUNCTION    DoKeyWord(KeywCode:LONGINT;Params:ParamType;ItemNr:LONGINT): LONGINT; override;
                     PROCEDURE   DoSection(SectionName:String;Level:LONGINT); OVERRIDE;
                     PROCEDURE   NonKeyword   (S:String); OVERRIDE;
                    END;


       TwoChar=ARRAY[0..1] OF CHAR;           {Used in class conversions}


{"Converts" v1.x category to v2.x, can also be used as shorthand}
CONST Cook1xto2xTable : ARRAY[65..90] OF TwoChar= ( 'AA','BB','CM','DD','EE','GF',
                                                    'WM','EH','II','EJ','CM','LL',
                                                    'GM','QN','QS','DP','QQ','DR',
                                                    'GS','ST','SM','VV','WW','DX',
                                                     'YY','ZZ');


CONSTRUCTOR TCookieConfig.NewConfigFile;

BEGIN
 Inherited NewConfigFile;
 CurKeys:=NIL;
 CurCateg:=NIL;
 FillChar(Configs,SIZEOF(Configs),#0);
 NEW(Configs[0]);
 FillChar(Configs[0]^,SIZEOF(ConfigRecord),#0);
 NrConfigs:=0;
 Configs[0]^.AllEnabled:=TRUE;
 Configs[0]^.ConfigName:='DEFAULT';
END;

PROCEDURE TCookieConfig.ConfigError(ErrNumber: LONGINT;Column:LONGINT;Data:String);

BEGIN
 CASE ErrNumber OF      {Errors known by CookieConfig are processed here,
                          unknown errors are passed to the TConfigFile handler}
  CookErr_ErrInBlockItem        : Writeln('Block or allow command with bad parameter "',Data,'" near line ',Line);
  CookErr_MoreConfigThanAllowed : WritelN('You defined more sections in the configurationfile than allowed (',MaxConfig,')');

 ELSE
  Inherited ConfigError(ErrNumber,Column,Data);
 END;
 HALT;
END;


FUNCTION TwoCharToCateg(Two: TwoChar):LONGINT;

BEGIN
 TwoCharToCateg:=ORD(Two[0])-65 + (ORD(Two[1])-65)*26;
END;

CONST Number = [#48..#57];
      Chars = [65..90,97..122];


FUNCTION DoBlock(VAR S : String;VAR Cat : CategSelectRec):LONGINT;

FUNCTION GetNumber(VAR Posi:LONGINT):LONGINT;
{Get decimal number at position thisbuf[Posi], update posi until first
non-decimal char, check for length}

VAR O : LONGINT;

BEGIN
 O := 0;
 IF NOT(S[Posi] IN Number) THEN
  EXIT(-1);
 WHILE (S[Posi] IN Number) AND (Posi<=Length(S)) DO
  BEGIN
   O := O*10+ORD(S[Posi])-48;
   INC(Posi);
  END;
 GetNumber := O;
END;

VAR
    A,B,C,D  : CARDINAL;
    CatHi,CatLo:LONGINT;

FUNCTION HandleNummericStuff:LONGINT;
{ Read something like -10 (0..10)  5-10 (5..10) or 6- (6..32767) }

BEGIN
  CASE S[C] OF
    '-' :      BEGIN
                INC(C);
                CatHi:=GetNumber(C);
               END;
   '0'..'9' : BEGIN
                CatLo:=GetNumber(C);
                IF C>Length(S) THEN
                 CatHi:=CatLo
                ELSE
                 BEGIN
                  IF S[C]='-' THEN
                   BEGIN
                    INC(C);
                    CatHi:=GetNumber(C);
                   END
                  ELSE
                    EXIT(CookErr_ErrInBlockItem);
                 END
               END
          ELSE  {CASE}
            EXIT(CookErr_ErrInBlockItem);
          END;
  IF (CatHi=-1) OR (CatLo=-1) THEN
   EXIT(CookErr_ErrInBlockItem);
  HandleNummericStuff:=0;
END;

BEGIN
 IF Length(S)=0 THEN
  EXIT(CookErr_ErrInBlockItem);
CatLo:=0; CatHi:=32767;
IF S[1] IN ['A'..'Z'] THEN
 BEGIN
  B:=ORD(S[1])-65;
  IF (S[2] IN ['A'..'Z']) AND (Length(S)>1) THEN
   BEGIN
    B:=B*26+ORD(S[2])-65;
    IF Length(S)>2 THEN
     BEGIN
      C:=3;
      D:=HandleNummericStuff;
      IF D<>0 THEN
        EXIT(D);
     END {Length>2}
    END
   ELSE {Single char semi v1.x compat aliases}
    BEGIN
     B:=TwoCharToCateg(Cook1xto2xTable[B]);
     IF Length(S)<>1 THEN
      BEGIN
       C:=2;
       D:=HandleNummericStuff;
      IF D<>0 THEN
        EXIT(D);
      END
    END;
  END
 ELSE
  EXIT(CookErr_ErrInBlockItem);
 Cat.Lower:=CatLo;
 Cat.Upper:=CatHi;
 Cat.Category:=B;
 DoBlock:=0;
END;

{Handles known keywords (ExampleKeywords array)}
FUNCTION TCookieConfig.DoKeyWord(KeywCode:LONGINT;Params:ParamType;ItemNr:LONGINT):LONGINT;


PROCEDURE CheckForEnlargeArray;

VAR AuxPtr : ^LONGINT;
    I      : LONGINT;
    Cat    : CategSelectRec;

BEGIN
 WITH Configs[NrConfigs]^ DO
  BEGIN
 IF (CatAllocated<>0) AND (CatUsed<>0) THEN
  INC(CatUsed);
 IF (CatAllocated=CatUsed) OR (CatAllocated=0) THEN
  BEGIN
   GetMem(AuxPtr,(GrowGap+CatAllocated)*SIZEOF(LONGINT));
   IF CatAllocated <> 0 THEN
    BEGIN
     Move (Categories[0],AuxPtr[0],CatAllocated*SIZEOF(LONGINT));
     Freemem(Categories,CatAllocated*SIZEOF(LONGINT));
    END;
   Categories:=AuxPtr;
   GetMem(AuxPtr,(GrowGap+CatAllocated)*SIZEOF(LONGINT));
   IF CatAllocated <> 0 THEN
    BEGIN
     Move (Subcats[0],AuxPtr[0],CatAllocated*SIZEOF(LONGINT));
     Freemem(Subcats,CatAllocated*SIZEOF(LONGINT));
    END;
   SubCats:=AuxPtr;
   INC(CatAllocated,GrowGap);
  END;
  Params.Value:=UPCASE(Params.Value);
  I:=DoBlock(Params.Value,Cat);
  IF I<>0 THEN
   ConfigError(I,0,Params.Value);
  Categories[CatUsed]:=Cat.Category;
  SubCats[CatUsed]:=Cat.Lower+Cat.Upper SHL 16;
  END;
END;



BEGIN
//  Writeln;
//  Write(Keyw[keywcode].name,' ':20-length(Keyw[keywcode].name));
  CASE CookieKeywEnum(KeywCode) OF
     Key_Block             : IF NOT Configs[NrConfigs]^.AllEnabled THEN
                              BEGIN
                               Writeln('AllDisabled used together with block which isn''t allowed in line ',Line);
                               HALT;
                              END
                             ELSE
                               CheckForEnlargeArray;
     Key_Allow             : IF Configs[NrConfigs]^.AllEnabled THEN
                              BEGIN
                               Writeln('AllEnabled used together with allow which isn''t allowed. (AllEnabled is');
                               Writeln('the default, until a AllEnabled or AllDisabled keyword is found) in line ',Line);
                               HALT;
                              END
                             ELSE
                               CheckForEnlargeArray;
     Key_DefCookie         : Configs[NrConfigs]^.DefCookie:=Params.Value;
     key_DefOffensive      : Configs[NrConfigs]^.DefOffensive:=Params.Value;
     key_Alldisabled           : BEGIN
                                  IF Configs[NrConfigs]^.EnableTriggered THEN
                                   BEGIN
                                    Writeln('Enabled or Disabled already used in this section which isn''t allowed ',Line);
                                    HALT;
                                   END;
                                  Configs[NrConfigs]^.EnableTriggered:=TRUE;
                                  Configs[NrConfigs]^.AllEnabled:=FALSE;
                                 END;
     key_AllEnabled            : BEGIN
                                  IF Configs[NrConfigs]^.EnableTriggered THEN
                                   BEGIN
                                    Writeln('Enabled or Disabled already used in this section which isn''t allowed  ',Line);
                                    HALT;
                                   END;
                                  Configs[NrConfigs]^.EnableTriggered:=TRUE;
                                  Configs[NrConfigs]^.AllEnabled:=TRUE;
                                 END;
 END;
END;

{Detection of an ini'sh section advances to the next configuration record}
PROCEDURE TCookieConfig.DoSection(SectionName:String;Level:LONGINT);

BEGIN
 IF (MaxConfig-1)>NrConfigs THEN
  BEGIN
   INC(NrConfigs);
   NEW(Configs[NrConfigs]);
   FillChar(Configs[NrConfigs]^,SIZEOF(ConfigRecord),#0);
   Configs[NrConfigs]^.AllEnabled:=TRUE;
   Configs[NrConfigs]^.ConfigName:=SectionName;
  END
 ELSE
  ConfigError(CookErr_MoreConfigThanAllowed,0,'');
END;

{Unknown keywords trigger an error}
PROCEDURE TCookieConfig.NonKeyword   (S:String);

BEGIN
 ConfigError(ConfErr_UnknownKeyword,0,S);
END;

Var CookieConfig : TCookieConfig;

BEGIN
 CookieConfig:=TCookieConfig.NewConfigFile;
 CookieConfig.SetKeywords(ExampleKeywords);
 CookieConfig.Commentset:=[';'];
 CookieConfig.ReadConfig('cook32.cfg');
 CookieConfig.KillConfigFile;
END.


