
{--------------------ASSEMBLER versions of ShortString procedures -----------}

{ First two faster replacements of RTL procedures}

{ 2-4 times as fast as RTL version}
FUNCTION Pos(CONST SubStr,S : String):LONGINT; ASSEMBLER;

ASM
                  mov     SubStr,%esi
                  movzbl  (%esi),%ebx
                  test    %ebx,%ebx
                  je      .LPosEnd
                  cmp     $1,%ebx
                  jne     .LRealPos
                  movl    8(%ebp),%eax
                  pushw   1(%eax)
                  movl    12(%ebp),%eax
                  pushl   %eax
                  call    charpos
//                  call    _EPASSTR$$_CHARPOS$STRING$CHAR
                  jmp     .LPosEnd
.LRealPos:
            inc     %esi
                  mov     S,%edi
                  movzbl  (%edi),%ecx
                  inc     %edi
                  inc     %ecx
                  movb    (%esi),%al   // Load first char of substr

.LPosLoop:
                  repne                // search for first char of substr
                   scasb
                  jne     .LPosNotFound
                  push    %ecx
                  push    %edi
                  push    %esi
                  xor     %edx,%edx
                  mov     %ebx,%ecx
                  dec     %edi
                  repe
                   cmpsb
                  je      .LPosEq
                  inc     %edx
.LPosEq:
                  pop     %esi
                  pop     %edi
                  pop     %ecx
                  test    %edx,%edx
                  jne     .LPosLoop
                  mov     S,%esi
                  sub     %esi,%edi
                  mov     %edi,%eax
                  decl    %eax
                  jmp     .LPosEnd
  .LPosNotFound:  xor     %eax,%eax
  .LPosEnd:
END['EAX','EBX','ECX','EDX','ESI','EDI'];

{ 4 times as fast as RTL version, more for pure appending (Index>Length(S))}
PROCEDURE insert(const source : shortstring;var s : shortstring;index : StrLenInt) ; ASSEMBLER;

ASM
             mov          index,%ebx            // load "index"
             test         %ebx,%ebx             // index=0?
             jne          .LIndexOK
             inc          %ebx                  // then index=1 (prepend)
.LIndexOK:
             mov          S    ,%esi
             movzbl       (%esi),%eax           // load Length(S)
             mov          source,%edi           // load Length(Source)
             movzbl       (%edi),%edx
             mov          %eax,%ecx
             addl         %edx,%ecx             // Set new length. (always L(S)+L(Source))
             movb         %cl,(%esi)
             cmp          %ebx,%eax
             jae          .LInsertIndexOk       // index<=Length(S)
             addl         %eax,%esi             // else append.
             inc          %esi                  // calc position beyond last char in S
             xchg         %edi,%esi             // edi=S[Length(S)+1)] esi=Source[0]
             inc          %esi                  // Source[1]
             mov          %edx,%ecx             // Length to %ecx
             cld
             rep
              movsb                             // Append
             jmp          .LInsertEnd
.LInsertIndexOk:
             add          %eax,%esi             // calc end of string
             subl         %ebx,%eax             // calc no of chars to be moved
             inc          %eax                  // chars to copy +1 (move char ON S[Index] too)

             mov          %esi,%edi             // %edi=S[Index]
             addl         %edx,%edi             // %edi=S[Index+Length(Source)]
             mov          %eax,%ecx             // load index register with chars to copy
             std
             rep                                // Make room
              movsb
             cld
             mov        S,%edi                  // Load S
             mov        Source,%esi             // load source
             inc        %esi                    // Source[1]
             mov        %edx,%ecx               // Length(Source) to index register
             addl       %ebx,%edi               // calc S[Index]
             rep
              movsb                             // insert/prepend.
.LInsertEnd:
END['EAX','EBX','ECX','EDX','ESI','EDI'];


{------- Assembler versions of non RTL procedures ---------}

FUNCTION RPos(CONST SubStr,S : String):LONGINT; ASSEMBLER;

ASM
                  mov     SubStr,%esi
                  movzbl  (%esi),%ebx
                  test    %ebx,%ebx
                  je      .LRPosEnd
                  cmp     $1,%ebx
                  jne     .LRealRPos
                  movl    8(%ebp),%eax
                  pushw   1(%eax)
                  movl    12(%ebp),%eax
                  pushl   %eax
                  call   rcharpos
//                  call    _EPASSTR$$_RCHARPOS$STRING$CHAR
                  jmp     .LRPosEnd
.LRealRPos:       add     %ebx,%esi
                  mov     S,%edi
                  movzbl  (%edi),%ecx
                  addl    %ecx,%edi     // Add 1 based length
                  inc     %ecx
                  movb    (%esi),%al   // Load last char of substr
                  std
.LRPosLoop:       repne
                   scasb
                  jne     .LRPosNotFound
                  push    %ecx
                  push    %edi
                  push    %esi
                  xor     %edx,%edx
                  mov     %ebx,%ecx
                  inc     %edi
                  repe
                   cmpsb
                  je      .LRPosEq
                  inc     %edx
.LRPosEq:         pop     %esi
                  pop     %edi
                  pop     %ecx
                  test    %edx,%edx
                  jne     .LRPosLoop
.LDirtyexit:      mov     S,%esi
                  sub     %esi,%edi
                  mov     %edi,%eax
                  incl    %eax
                  incl    %eax
                  subl    %ebx,%eax
                  jmp     .LRPosEnd

  .LRPosNotFound: xor     %eax,%eax
  .LRPosEnd:      cld
END['EAX','EBX','ECX','EDX','ESI','EDI'];


PROCEDURE Invert (VAR S : String); ASSEMBLER;
ASM
                     mov    S,%esi
                     movzbl (%esi),%eax
                     test   %eax,%eax
                     je     .LInverEnd

                     mov    %esi,%edi
                     inc    %esi
                     addl   %eax,%edi
                     shr    $1,%eax
                     mov    %eax,%ecx
.LinvertLoop:        mov    (%esi),%al
                     movb   (%edi),%bl
                     movb   %al,(%edi)
                     mov    %bl,(%esi)
                     incl   %esi
                     decl   %edi
                     loop  .LinvertLoop
.LInverEnd:
end['EAX','EBX','ECX','ESI','EDI'];

{------- The real, the only, the ultimate; assembler ---------------}

PROCEDURE LTrim(VAR P : String;Ch:CHAR);    ASSEMBLER;
  asm
                   movl P,%esi       // Load String
                   xor  %eax,%eax
                   lodsb
                   movl %esi,%edi
                   movl %eax,%ecx
                   movl %ecx,%ebx
                   movb Ch,%al      // Load character
                   cld
                   repe
                    scasb                // Search for NOT equal, no longer
                                         // then string length
                   dec %edi
                   cmpl %edi,%esi        // Chars found?
                   jae   .LNothingToDo
                   movl %esi,%ecx        // ecx=esi-edi Result is negative
                   subl %edi,%ecx        // add ebx


                   addl %ebx,%ecx        // Chars to be copied calculated
                   mov  %ecx,%ebx        // and saved.
                   xchgl %esi,%edi
  .LTRIM1:         lodsb
                   stosb
                   decl  %ecx
                   jnz .LTRIM1
                   mov  P,%edi     // update length
                   movb %bl,(%edi)
  .LNothingToDo:
end ['EAX','EBX','ECX','ESI','EDI'];

PROCEDURE KillChar(VAR S : STRING;CONST CSet:CHARSET); ASSEMBLER;

  asm

                   movl S,%edi       // Load String
                   movl %edi,%esi
                   xor  %eax,%eax
                   lodsb

                   mov   %eax,%ecx
                   movl  %ecx,%ebx
                   mov   CSet,%edi

  .LKILLCHAR4:
                   jecxz    .LNothingToDo3

                   lodsb
                   mov   %eax,%edx
                   andl  $0xf8,%edx
                   shrl  $3,%edx
                   andl  $7,%eax
                   decl  %ecx
                   btl   %eax,(%edi,%edx)
                   jc   .LKILLCHAR4
     .LKILLCHAR3:  dec  %esi
                   movl S,%edi       // Load String
                   inc  %edi
                   cmpl %edi,%esi        // Chars found?
                   je   .LNothingToDo2
                   movl %edi,%ecx        // ecx=esi-edi Result is negative
                   subl %esi,%ecx        // add ebx
                   addl %ebx,%ecx        // Chars to be copied calculated
                   mov  %cl,-1(%edi)     // New length saved
                   rep
                    movsb
 .LNothingToDo2:   leave
                   ret $8

.LNothingToDo3:   mov S ,%edi
                  movb $0,(%edi)
end ['EAX','EBX','ECX','EDX','ESI','EDI'];

{ Old Basic procedure String$, fill string with chars.}
PROCEDURE StrStr(VAR P:String;C:Char;Count:LONGINT); ASSEMBLER;

asm

                   mov   P,%edi
                   cld
                   mov   Count,%ecx
                   mov   %ecx,%eax
                   stosb
                   movl  C,%eax
                   movb  %al,%ah
                   movl  %eax,%ebx
                   shll  $16,%eax
                   movw  %bx,%ax
                   movl  %ecx,%ebx
                   andl  $3,%ebx
                   shrl  $2,%ecx
                   rep
                    stosl
                   movl  %ebx,%ecx
                   rep
                    stosb
end ['EAX','EBX','ECX','EDI'];

PROCEDURE AppendBackslash(VAR S : String);  ASSEMBLER;
      ASM

                         movl S,%esi
                         mov %esi,%edi
                         cld
                         xor %eax,%eax
                         lodsb
                         add %eax,%esi
        // Now we are pointing to the last char+one.

{$IFDEF Linux}
                         cmpb $47,-1(%esi)
                         je .Lappendbackslashready
                         movb $47,(%esi)      // append backslash
{$ELSE}
                         cmpb $92,-1(%esi)
                         je .Lappendbackslashready
                         movb $92,(%esi)      // append backslash
{$ENDIF}
                         incb (%edi)          // update length
.Lappendbackslashready:
END ['EAX','ESI','EDI'];

PROCEDURE RTrim(var P:String;Ch:Char); ASSEMBLER;
  asm

                   movl P,%edi       // Load String
                   cld
                   movzbl (%edi),%eax
                   add %eax,%edi
                   mov %eax,%ecx
                   movb Ch,%al
                   std
                   repe
                    scasb
                   cld
                   incl %ecx
                   mov P,%edi
                   movb %cl,(%edi)
end ['EAX','ECX','EDI'];

PROCEDURE KillBChar(VAR S : String;const CSet:CharSet); ASSEMBLER;
  asm

                   movl S,%edi       // Load String
                   xor  %eax,%eax
                   movb (%edi),%al
                   add  %eax,%edi
                   movl %eax,%ecx
                   jecxz .LNoCharsleft
                   movl  CSet,%esi
                   incl  %ecx
.LKillBCharLoop:   jecxz .LKillBCharTrimit
                   movb  (%edi),%al
                   movl  %eax,%edx            // And clears the high bytes
                   andl  $0xf8,%edx
                   shrl  $3,%edx
                   andl  $7,%eax
                   decl  %ecx
                   je    .LNoCharsleft
                   decl  %edi

                   btl   %eax,(%esi,%edx)
                   jc   .LKillBCharLoop

.LKillBCharTrimit:
.LNoCharsleft:     mov  S,%esi
                   movb %cl,(%esi)
end ['EAX','EBX','ECX','EDX','ESI','EDI'];

(* Replace in string S the character "ReplaceMe" with "RepWith".*)
PROCEDURE ReplaceChar(VAR S : String;ReplaceMe,RepWith:CHAR); ASSEMBLER;
ASM

                      movl   S,%esi       // Load String
                      cld
                      xorl   %eax,%eax
                      lodsb
                      movl   %eax,%ecx
                      jecxz  .LReplaceCharend
                      movl   ReplaceMe,%ebx
                      movl   RepWith,%eax
.LReplaceCharLoop:    cmpb   %bl,(%esi)
                      jne    .LReplaceCharNextChar
                      movb   %al,(%esi)
.LReplaceCharNextChar: incl   %esi
                      loop   .LReplaceCharLoop
.LReplaceCharend:
end ['EAX','EBX','ECX','ESI'];

PROCEDURE RGrow(VAR S : String;C:CHAR;Count:LONGINT); ASSEMBLER;

ASM

                      movl S,%edi            // Load String
                      movzbl (%edi),%eax            // Load length
                      movl Count,%ecx         // load count
                      cmp  %ecx,%eax               // subtract size
                      jae  .LRStringBigEnough
                      mov  %cl,(%edi)              // store new size
                      addl %eax,%edi
                      inc  %edi
                      sub  %eax,%ecx
                      movb C,%al
                      rep
                       stosb                      //  store padding
.LRStringBigEnough:
end ['EAX','ECX','EDI'];

PROCEDURE LGrow(VAR S: STRING;Ch:CHAR;Count:LONGINT); ASSEMBLER;

ASM

                      movl S,%edi       // Load String
                      movzbl (%edi),%ecx
                      movl   %ecx,%ebx
                      mov  Count,%eax     // load count
                      cmp  %eax,%ebx           // compare size
                      ja   .LGrowStringBigEnough
                      mov  %al,(%edi)          // length
                      sub  %ebx,%eax
                      mov  %eax,%ebx           // bx is nr pads, cx is length
                      addl %ecx,%edi
                      mov  %edi,%esi
                      addl %ebx,%edi           // moving up, std required,
                      std
                      rep                      // move string, make room
                       movsb                   //  for padding
                      cld
                      movb Ch,%al
                      movl S,%edi        //  reLoad String and pad-char
                      inc %edi
                      movl %ebx,%ecx           //  count padding chars
                      rep
                      stosb
                      mov  %ebx,%eax
.LGrowStringBigEnough:

end ['EAX','EBX','ECX','ESI','EDI'];

{Remove C entirely from P}
PROCEDURE StripChar(VAR S:String;C:CHAR); ASSEMBLER;
   ASM
                           mov   S,%esi
                           xor   %eax,%eax
                           lodsb
                           mov   %esi,%edi
                           mov   %eax,%ecx
                           mov   %esi,%edx
                           jecxz .LStripcharNoChar
                           movb  C,%bl
        .LStripCharLoop:   lodsb
                           cmpb  %al,%bl
                           je    .LStripCharSkip
                           stosb
        .LStripCharSkip:   loop .LStripCharLoop
                           sub  %edx,%edi
                           xchg %edx,%edi
                           movb  %dl,-1(%edi)
        .LStripcharNoChar:
END ['EAX','EBX','ECX','EDX','ESI','EDI'];


{StripChar for a CHARSET; Remove any char in C entirely from P}
PROCEDURE KillChrTot(VAR S : String;CONST CSet:CHARSET); ASSEMBLER;
       ASM

                 movl  S,%esi
                 xor   %eax,%eax
                 lodsb
                 mov   %esi,%edi
                 mov   %eax,%ecx
                 mov   CSet,%ebx
                 jecxz .LKillChrTotNoChar
.LKillChrTotLoop:lodsb
                 pushl %eax
                 mov   %eax,%edx
                 andl  $0xf8,%edx
                 shrl  $3,%edx
                 andl  $7,%eax
                 btl   %eax,(%ebx,%edx)
                 jc    .LKillChrTotSkip
                 popl  %eax
                 stosb
.LKillChrTotSkip:  loop .LKillChrTotLoop
                 mov  S,%edx
                 sub  %edx,%edi
                 decl %edi
                 movl %edi,%eax
                 movb %al,(%edx)
.LKillChrTotNoChar:
END ['EAX','EBX','ECX','EDX','ESI','EDI'];


{ Pos for one char only. Faster than an ordinary Pos}
FUNCTION CharPos(CONST S :String;C:Char):LONGINT; ASSEMBLER;

    ASM // Pas strings allow repne scasb, probably very fast

                      movl  S,%edi       // Load String
                      cld
                      movzbl   (%edi),%ecx
                      inc   %edi
                      inc  %ecx
                      mov   %ecx,%ebx
                      movb  C,%al
                      repne
                       scasb
                      jne  .Lcharposnotfound
                      sub  %ecx,%ebx
                      mov  %ebx,%eax
                      leave
                      ret $6
.Lcharposnotfound:    xor %eax,%eax
END ['EAX','EBX','ECX','EDI'];

{ Like CharPos but starts searching on character number Count (zero based),
    returnvalue is zero based and absolute (to beginning of string)}
FUNCTION NextCharPos(CONST S:String;C:Char;Count:LONGINT):LONGINT; ASSEMBLER;

    ASM // Pas strings allow repne scasb, probably very fast

                      movl  S,%edi       // Load String
                      movl Count,%edx
                      cld
                      movzbl (%edi),%ecx
                      cmpl %ecx,%edx
                      ja .LNcharposnotfound
                      addl %edx,%edi              // Start on higher address
                      subl %edx,%ecx
                      mov   %ecx,%ebx
                      inc  %ecx
                      movb  C,%al
                      repne
                       scasb
                      jne  .LNcharposnotfound
                      sub  %ecx,%ebx
                      mov  %ebx,%eax
                      addl %edx,%eax
                      leave
                      ret $10
.LNcharposnotfound:   xor %eax,%eax
END ['EAX','EBX','ECX','EDX','EDI'];

PROCEDURE StripDoubleChar(VAR S:String;C:Char); ASSEMBLER;

        ASM

                          movl   S,%esi       // Load String
                          xor    %eax,%eax
                          lodsb
                          movl   %esi,%edi
                          movl   %esi,%edx
                          mov    %eax,%ecx
                          movb   C,%bl
  .LStripDoubleLoop:      lodsb
                          cmp    %al,%bl          // Check for right char
                          jne    .LStripDoubleCheck
                          cmpb   %al,(%esi)       // Equal to next, skip
                          je     .LStripSkipChar
.LStripDoubleCheck:       stosb
.LStripSkipChar:          loop .LStripDoubleLoop
.LStripDoubleExit:        sub  %edx,%edi
                          xchg %edx,%edi
                          movb %dl,-1(%edi)

END ['EAX','EBX','ECX','EDX','ESI','EDI'];

FUNCTION CharPosSet(CONST S : String;CONST CSet:CHARSET):LONGINT; ASSEMBLER;

    ASM

                      movl  S,%esi       // Load String
                      movzbl (%esi),%ecx

                      mov   %ecx,%edi          // Keep backup.
                      incl  %esi
                      movl  CSet,%ebx
.LCharPosSetLoop:     lodsb
                      movl  %eax,%edx
                      andl  $0xf8,%edx
                      shrl  $3,%edx
                      andl  $7,%eax
                      dec   %ecx
                      js    .LCharPosSetEOS
                      btl   %eax,(%ebx,%edx)
                      jnc   .LCharPosSetLoop
                      sub   %ecx,%edi
                      mov   %edi,%eax
                      leave
                      ret $8
.LCharPosSetEOS:      xor %eax,%eax

END ['EAX','EBX','ECX','EDX','ESI','EDI'];

FUNCTION RCharPos(CONST S :String;C:Char):LONGINT; ASSEMBLER;

    ASM

                      movl S,%edi       // Load String
                      xor  %eax,%eax
                      movb (%edi),%al
                      add  %eax,%edi
                      mov  %eax,%ecx
                      inc  %ecx
                      movb C,%al
                      std
                      repne
                       scasb
                      jne  .Lrcharposnotfound
                      cld
                      mov %ecx,%eax
                      leave
                      ret $6
.Lrcharposnotfound:   xor %eax,%eax
END ['EAX','ECX','EDI'];

FUNCTION NextRCharPos(CONST S: String;C:Char;Count:LONGINT):LONGINT; ASSEMBLER;

    ASM

                          movl   S,%edi       // Load String
                          movzbl (%edi),%eax
                          movl  Count,%ecx
                          cmp  %eax,%ecx            // Count>Length
                          ja   .Lnrcharposnotfound2
                          add  %ecx,%edi
                          inc  %ecx
                          movb C,%al
                          std
                          repne
                           scasb
                          jne  .Lnrcharposnotfound2
                          cld
                          mov %ecx,%eax
                          leave
                          ret $10
    .Lnrcharposnotfound2:  xor %eax,%eax
END ['EAX','ECX','EDI'];

FUNCTION SCompare(CONST S,S2:String):LONGINT;ASSEMBLER;

    ASM

                          movl   S,%esi       // Load String 1
                          movl  S2,%edi       // Load String 2
                          movzbl (%esi),%eax
                          movzbl (%edi),%ebx
                          incl  %esi
                          incl  %edi
                          mov   %eax,%ecx
                          cmp   %eax,%ebx
                          ja  .LEBXbiggest
                          mov   %ebx,%ecx        // smallest size in EBX.

.LEBXbiggest:             jecxz .LTestLength
                          repe
                           cmpsb
                          ja .LAbove
                          jb .LBelow
.LTestLength:             cmp %eax,%ebx
                          jb  .LAbove
                          ja .LBelow
                          xor %eax,%eax
                          leave
                          ret $8
.LBelow:                  movl $-1,%eax
                          leave
                          ret $8
.LAbove:                  movl $1,%eax
.LSaveexit:
END['EAX','EBX','ECX','ESI','EDI'];

PROCEDURE UpperCase(VAR S : String); ASSEMBLER;
       ASM
                   movl  S,%esi
                   movzbl (%esi),%ecx
                   jecxz .LUppernothingtodo
                   incl %esi
                   mov %esi,%edi
.Lupperloop:       lodsb
                   cmp  $97,%al
                    jb  .LUpperNotthischar
                    cmp  $122,%al
                    ja  .LUpperNotthischar
                    subb $32,%al
.LUpperNotthischar: stosb
                    loop .Lupperloop
.LUppernothingtodo:
END ['EAX','ECX','ESI','EDI'];

PROCEDURE LowerCase(VAR S : String); ASSEMBLER;
       ASM
                   movl  S,%esi
                   movzbl (%esi),%ecx
                   jecxz .LLowernothingtodo
                   incl %esi
                   mov %esi,%edi
.LLowerloop:       lodsb
                   cmp  $65,%al
                    jb  .LLowerNotthischar
                    cmp  $90,%al
                    ja  .LLowerNotthischar
                    addb $32,%al
.LLowerNotthischar: stosb
                    loop .LLowerloop
.LLowernothingtodo:
END ['EAX','ECX','ESI','EDI'];

PROCEDURE Commastr(var S : String;sep:CHAR); ASSEMBLER;

ASM
          mov   S,%edi
          xor   %eax,%eax
          movb  (%edi),%al
          mov   %eax,%ebx
          mov   $3,%ecx
          cltd // cwde
          idiv  %ecx
          je   .LLessthan4
          test  %edx,%edx
          jne   .LNotDiv3
          dec   %eax
          je  .LLessthan4
.LNotDiv3:
          add   %eax,(%edi)
          add   %ebx,%edi
          mov   %edi,%esi
          add   %eax,%edi
          mov   %eax,%ecx
          movzbl sep,%eax
          std
.LNext3:
          movsb
          movsb
          movsb
          stosb
          loop .LNext3
          cld
.LLessthan4:
END['EAX','EBX','ECX','EDX','ESI','EDI'];

PROCEDURE ExpandTabs(CONST P : String;VAR P2:String;Tabsize:LONGINT); ASSEMBLER;

ASM
            movl  P,%esi       // Load PChar
            movzbl (%esi),%ecx
            jecxz .LExpandEmpty
            movl  P2,%edi
            inc  %esi
            inc  %edi
            cld
            xor %ebx,%ebx
.LPExpandLoop:lodsb
            cmp $9,%al
            je .LPExpandThisTab
            inc %ebx
            stosb
.LExpandRet:loop .LPExpandLoop
            jmp .LPExpandReady
.LPExpandThisTab:
            push %ecx
            movl Tabsize,%ecx
            mov %ebx,%eax
//            cwde
            cltd
            idiv %ecx
            subl %edx,%ecx
            addl %ecx,%ebx
            movb $32,%al
            rep
             stosb
            pop %ecx
            jmp .LExpandRet
.LPExpandReady:
            mov P2,%ecx
            mov %ecx,%esi           //0123456
            sub %ecx,%edi
            dec %edi
            mov %edi,%ecx
            movb %cl,(%esi)
.LExpandEmpty:
end['EAX','EBX','ECX','EDX','ESI','EDI'];

PROCEDURE BinaryToStr(VAR S : String;Value,Bits : CARDINAL);ASSEMBLER;

ASM
                            movl    S,%edi
                            movl    Value,%ebx
                            movl    Bits,%edx
                            movb    %dl,(%edi)      // Save length
                            mov     $32,%ecx
                            sub     %edx,%ecx        // calc unwanted bits
                            jecxz   .LBinaryNoShl    // 32-Bits, test for 0
                            shll    %cl,%ebx         // Shift left to remove
                                                     // unwanted bits.
.LBinaryNoShl:
                            mov     %edx,%ecx        // Number of bits to do
                            jecxz   .LBinaryNothing  // test for no-bits.
                            inc     %edi             // skip length

.LBinaryLoop:               xor     %eax,%eax        // Empty eax
                            shl     $1,%ebx          // highest bit to carry
                            adc     $48,%eax         // add carry to ascii 0
                            stosb                    // save in string
                            loop    .LBinaryLoop
.LBinaryNothing:
end ['EAX','EBX','ECX','EDX','EDI'];

FUNCTION StrToBinary(CONST S : String;Bits : CARDINAL):CARDINAL; ASSEMBLER;

ASM
                            xor     %ebx,%ebx     // start with empty value.
                            movl    S,%esi
                            movl    bits,%ecx
                            jecxz   .LStrToBinNothing
                            inc     %esi          // skip length.

.LStrToBinNextbit:          shl     $1,%ebx
                            lodsb
                            and     $1,%eax       // 48 and 1=0 | 49 and 1=1
                            or      %eax,%ebx     // move into low bit
                            loop    .LStrToBinNextbit
.LStrToBinNothing:          mov     %ebx,%eax     // set als return value.
end['EAX','EBX','ECX','ESI'];

FUNCTION StrToHex(CONST S : String;Digits : CARDINAL):CARDINAL; ASSEMBLER;

ASM
                            xor     %ebx,%ebx     // start with empty value.
                            movl    S,%esi
                            movl    Digits,%ecx
                            jecxz   .LStrToHexNothing
                            inc     %esi          // skip length.

.LStrToHexNextbit:          shl     $4,%ebx
                            xor     %eax,%eax
                            lodsb
                            subl    $48,%eax
                            cmp     $10,%eax
                            jb      .LHexlow
                            subl    $7,%eax
                            and     $15,%eax       // protection against bad
                                                   // digit parameters.
        .LHexlow:           addl    %eax,%ebx
                            loop    .LStrToHexNextbit
.LStrToHexNothing:          mov     %ebx,%eax     // set als return value.
end['EAX','EBX','ECX','ESI'];

FUNCTION StrToOct(CONST S : String;Digits : CARDINAL):CARDINAL; ASSEMBLER;
ASM
                            xor     %ebx,%ebx     // start with empty value.
                            movl    S,%esi
                            movl    Digits,%ecx
                            jecxz   .LStrToOctNothing
                            inc     %esi          // skip length.

.LStrToOctNextbit:          shl     $3,%ebx
                            xor     %eax,%eax
                            lodsb
                            subl    $48,%eax
                            and     $7,%eax       // protection against bad
                                                   // digit parameters.
        .LOctlow:           addl    %eax,%ebx
                            loop    .LStrToOctNextbit
.LStrToOctNothing:          mov     %ebx,%eax     // set als return value.
end['EAX','EBX','ECX','ESI'];

PROCEDURE HexToStr(VAR S : String;Value,Digits : CARDINAL);ASSEMBLER;
ASM
                            movl    S,%edi

                            movl    Digits,%edx
                            mov     %edx,%ebx
                            shll    $2,%ebx
                            movb    %dl,(%edi)       // Save length
                            mov     $32,%ecx
                            sub     %ebx,%ecx        // calc unwanted bits
                            movl    Value,%ebx
                            jecxz   .LHexNoShl       // 32-Digits, test for 0
                            shll    %cl,%ebx         // Shift left to remove
                                                     // unwanted Digits.
.LHexNoShl:                 mov     %edx,%ecx
                            inc     %edi
.LHexLoop:                  mov     %ebx,%eax
                            shr     $28,%eax
                            addl    $48,%eax
                            cmp     $58,%eax
                            jb      .LHexStore
                            addl    $7,%eax
.LHexStore:                 stosb
                            shl     $4,%ebx
                            loop    .LHexLoop
.LHexNothing:
end ['EAX','EBX','ECX','EDX','EDI'];

PROCEDURE OctToStr(VAR S : String;Value,Digits : CARDINAL);ASSEMBLER;
ASM
                            movl    S,%edi

                            movl    Digits,%edx
                            mov     %edx,%ebx
                            shll    $1,%ebx          // bits*3
                            addl    %edx,%ebx
                            movb    %dl,(%edi)       // Save length
                            mov     $32,%ecx
                            sub     %ebx,%ecx        // calc unwanted bits
                            movl    Value,%ebx
                            jecxz   .LOctNoShl       // 32-Digits, test for 0
                            shll    %cl,%ebx         // Shift left to remove
                                                     // unwanted Digits.
.LOctNoShl:                 mov     %edx,%ecx
                            inc     %edi
.LOctLoop:                  mov     %ebx,%eax
                            shr     $29,%eax
                            addl    $48,%eax
                            stosb
                            shl     $3,%ebx
                            loop    .LOctLoop
.LOctNothing:
end ['EAX','EBX','ECX','EDX','EDI'];
