UNIT EPasStr;
{Extra Strings routines, not found in the normal FPC RTL. (Some exist for
    AnsiStrings in SysUtils)
Pascal String[] + Overloaded AnsiStrings version.

   This is a unit of the XTDLIB/FPC package
   COPYRIGHT (C) 1993..1998 by Marco van de Voort,  All Rights Reserved
   Free for non-commercial use. See license.txt.
   (www.stack.nl/~marcov/xtdlib.htm)
   Free for non-commercial use.

 Some are basic stringprocedures (like the LTrim, RTrim, Kill* procedure and
  the simpler CharPos procedures). Some are really weird, like StripDoubleChar
  and GetBetween(the last one not ported yet)

 The routines all passed a thorough test, and are fairly stable. I however
 incidentally find bugs, so it's possible that small bugs exist, specially in
 the more complicated procedures.
 Please report them, so that future versions will be entire stable.

 Procedures marked with PAS don't exist in assembler,
 For ANSISTRING only CharPos and variants were optimized with assembler,
 except NextCharPosSet which is now pascal in both variants.

 This is a HUGE unit, about 20k .O  May I suggest smartlinking :-)

 Some routines have equivalents in SYSUTILS. I don't use SYSUTILS though, since
 it forces Delphi and STRING=ANSISTRING ($H+) mode.

 }

INTERFACE
{$DEFINE Useasm}                {Default : Use assembler, comment this off to
                                         compile pascal code, automatically
                                        shutdown if you UNDEF I386 in syseq.inc}

{$DEFINE AnsiUseAsm}            {Assembler for some Ansistring routines.
                                  When AnsiString isn't alpha anymore, this
                                  will be global with UseAsm}

{$I syseq.inc}                  { standard defines used for XTDLIB}

{$Ifdef FPC}
{$ifdef Useasm}
{$Info EPasStr : Using assembler routines when possible }
{$else}
{$Info EPasStr : Using All-Pascal routines }
{$endif Useasm}
{$ENDIF}
{$IFDEF Turbo}
 TYPE CARDINAL=LONGINT;
{$ENDIF}

TYPE CHARSET = SET OF CHAR;

{---- Procedures which remove/replace characters in the String ---}

{Strip character Ch from the left of the string }
PROCEDURE LTrim (VAR P : String;Ch:CHAR);

{ More powerful LTrim. Strip al characters in Ch from beginning of string}
PROCEDURE KillChar(VAR S : STRING;CONST CSet:CHARSET);

{Strip character Ch from the right (end) of the string}
PROCEDURE RTrim(VAR P:String;Ch:Char);

{ More powerful RTrim. Strip al characters in Ch from ending of string}
PROCEDURE KillBChar(VAR S : String;CONST CSet:CHARSET);

{ Replace in string S the character "ReplaceMe" with "RepWith".}
PROCEDURE ReplaceChar(VAR S : String;ReplaceMe,ReplaceWith:CHAR);

{Remove C entirely from P}
PROCEDURE StripChar(VAR S:String;C:CHAR);

{StripChar for a CHARSET; Remove any char in C entirely from P}
PROCEDURE KillChrTot(VAR S : String;CONST CSet:CHARSET);

{If 2 or more sequential 'C' chars exist in string, strip all but one
 ' 1   2 3   5  ' becomes ' 1 2 3 5 '
Used to make mail from Fido-newbies readable (run it for ' ','.' and '!') :-)}
PROCEDURE StripDoubleChar(VAR S:String;C:Char);

{ Invert string (write backwards) }
PROCEDURE Invert( VAR s :String);

{---- Pos procedures}

{ Pos variants for one char only. Faster than an ordinary Pos, but only a real
big difference for big strings else 33% faster or so.

All Pos variants use Scasb instruction, except CharPosSet
they return 0 when character not found }
FUNCTION CharPos(CONST S :String;C:Char):LONGINT;

{As CharPos, but start searching on Count}
FUNCTION NextCharPos(CONST S:String;C:Char;Count:LONGINT):LONGINT;

{ Same as CharPos, but starts searching at back of string}
FUNCTION RCharPos(CONST S :String;C:Char):LONGINT;

{ Same as CharPosl, but starts searching from position Count in the direction
    of the beginning of the string}
FUNCTION NextRCharPos(CONST S:String;C:Char;Count:LONGINT):LONGINT;

{ Replace first occurance of Sub1 scanning from the begin of string SS with
Sub2}
FUNCTION  Replace(Sub1,Sub2: String;SS: String ): String;

{ Replace first occurance of Sub1 scanning from the end of string SS with
Sub2}
FUNCTION ReplaceLast(Sub1,Sub2: STRING;VAR SS: STRING ): STRING;


{ CharPos, for a set}
FUNCTION CharPosSet(CONST S : String;CONST CSet:CHARSET):LONGINT;

{Next variant of CharPosSet, find first character in set C, start searching
at character position Count}
FUNCTION NextCharPosSet(CONST S : String;CONST C:CHARSET;Count:LONGINT):LONGINT;

{Like Pos(). Returns 0 when substr not found, otherwise standard STRING index
(1..Length(X)) to first char of last substr found in S)}
FUNCTION RPos(CONST SubStr,S : String):LONGINT;

{All (normal, not international ones) characters lowercase}
PROCEDURE LowerCase(VAR S : String);

{All (normal, not international ones) characters uppercase}
PROCEDURE UpperCase(VAR S : String);

{ Get first "bits" bits or digits from S (like S:='0101010'),
and return their binary value.
Allowable range StrToBinary parameter Bits  range is 0..32,
                StrToOct    parameter Digit range is 0..11,
                StrToHex    parameter Digit range is 0..8,
                though 0 is useless (returns 0)}

FUNCTION StrToBinary(CONST S : String;Bits  : CARDINAL):CARDINAL;
FUNCTION StrToOct   (CONST S : String;Digits: CARDINAL):CARDINAL;
FUNCTION StrToHex   (CONST S : String;Digits: CARDINAL):CARDINAL;
{------------------------------------------------------------------------

  Following procedures (could) write more characters to the string than
  when entering the procedure. Make sure the string is big enough
  This is a problem with all procedures enlarging a pascal string. Equivalents
  in SWAG always will have the same problem, it's inherent to the pascal string
  passing conventions (alloc-size not passed)
  }

{ If Length(S)<Count then pad end string with character C until
    Length(S)=Count}
PROCEDURE RGrow(VAR S : String;C:CHAR;Count:LONGINT);

{ If Length(S)<Count then pad begin string with character C until
    Length(S)=Count}
PROCEDURE LGrow(VAR S: STRING;Ch:CHAR;Count:LONGINT);

{Uses LGrow and RGrow to center string S. Pad char C, center width "ScreenWidth"}
Procedure CenterString(VAR S:String;C:Char;ScreenWidth:LONGINT);


{ Dirsearching primitive, adds a backslash to end of string if string doesn't
    end with a backslash. Uses forward slash if compiled for Linux version.}
PROCEDURE AppendBackslash(VAR S : String);

{ PAS,OVERLOADED
    Returns Nth substring of S in R (first is 0), separator is character T
One of my favourite string procedures. Multiple separators between char's
allowed}
{$IFDEF Turbo}
PROCEDURE ItemC(VAR R: STRING; CONST S: STRING; T: CHAR; N: LONGINT);
{$ELSE}
PROCEDURE Item(VAR R: STRING; CONST S: STRING; T: CHAR; N: LONGINT);
{$ENDIF}

{ PAS,OVERLOADED
     Returns Nth substring of S in R (first is 0), separators are all
characters in S. One of my favourite string procedures.
(Same as Item, but with more possible separators)}

PROCEDURE Item(VAR R: STRING; CONST S: STRING; CONST T: CHARSET; N: LONGINT);

{ PAS
  Returns Nth substring of S in R (first is 0), separator are characters in
    string T.
 Interface procedure, which converts string T to a CHARSET and calls Item.}
PROCEDURE ItemS(VAR R: STRING; CONST S: STRING; CONST T: String; N: LONGINT);

{ Same as itemish procedures. (returns Nth substring of S in R (first is 0),
separators are all characters in S). Supports several kinds of quoted strings.

However depending on the mode constant this procedure keeps track of quotes.

Mode=0   quote char is "
Mode=1   quote char is '
Mode=2   Modula-2 string mode. both " and ' work, and ' can be inside "" and
         vice versa (for strings as the following line (separator =space and
         comma) item1,"item 2 'still item2'" 'item3"nonsense" still item 3')

Return= 0  - all ok.
Return=-1 - invalid mode value
Return= other values -  the position of an unmatched quote

This function works, but quotes blocks without separators in between (
like aa'bb') aren't treated properly yet.  "AA""BB" works however.
NOT yet in help, NEW in 0.19
}
FUNCTION QItem(VAR R: STRING; CONST S: STRING; T: CHARSET; N: LONGINT;Mode : LONGINT):LONGINT;

{PAS, return the string between the first occurance of C1 or C2.
    If C1=C2, the string between first and second occurance of C1
    Returns status (existance of both characters and position C1< C2)}
FUNCTION GetBetween(Source:String;VAR Dest:String;C1,C2:CHAR):BOOLEAN;

{Convert Number(Value) to string(S) in binary, octal or hexadecimal
representation, with a configurable number of bits or digits,
Range of Bits  in BinaryStr  0..32,
Range of Digit in OctToStr   0..11
Range of Digit in HexToStr   0..8
Though value zero has no real meaning.}

PROCEDURE BinaryToStr(VAR S : String;Value,Bits   : CARDINAL);
PROCEDURE HexToStr   (VAR S : String;Value,Digits : CARDINAL);
PROCEDURE OctToStr   (VAR S : String;Value,Digits : CARDINAL);

{ Inserts character on each 3rd spot starting from the back.
(2123456789 -> 2,123,456,789}
PROCEDURE Commastr(var S : String;sep:CHAR);

{ Expand all tabs to spaces, TabSize is usually 8}
PROCEDURE ExpandTabs(CONST P : String;VAR P2:String;Tabsize:LONGINT);

{ Expand spaces to tabs (if possible), TabSize is usually 8}
PROCEDURE Compresstabs(Source:String;VAR Dest:String;Tabsize:LONGINT);

{$IFDEF Useasm}
{EXPERIMENTAL, non functional}
FUNCTION SCompare(CONST S,S2:String):LONGINT;
{$ENDIF}

{Insert for a character.
  Should be an Insert overloaded procedure. I can't since Insert is in a
    different unit}
PROCEDURE InsertC(VAR S : String;C:CHAR;Position:LONGINT);

{Old Basic functions}
FUNCTION MidStr( CONST StrName: String;  StartPos, NumChars : Integer) : String;
FUNCTION LeftStr( CONST StrName: String; NumChars : Integer) : String;
FUNCTION RightStr ( CONST StrName: String; NumChars : Integer) : String;
{ Old Basic procedure String$, fill string with chars.}
PROCEDURE StrStr(VAR P:String;C:Char;Count:LONGINT);

{ String matching procedure (any combination of * and ?) I think it is fast,
 since the source resembles the quite fast TopSpeed procedures (the actual
 procedure came from SWAG)}
function Match(CONST source, pattern: String): Boolean;

{ Modula2 equivalent of Copy, but 1-based offset (Contrary to M2, so basically
    copy in procedure form)}
PROCEDURE Slice (VAR R: String; S: String; P,L: LONGINT);

{ The ordinary Pos is horribly slow. I timed it, and even with relatively small
  strings it is at least twice as slow as an assembler variant (same algoritm,
  but use of string-instructions). For bigger (>20 chars) the RTL procedure is
  about 4 times as slow

  The assembler procedure is below, and only exists when useasm=true.

  Same for Insert which is about 4 times faster, the algoritm of Insert is
  slightly different but equal. Moves string only once for append situation,
  two times for real Insert.

  }

{$IFDEF sadjgUseAsm}
 FUNCTION Pos(CONST SubStr,S : String):LONGINT;
{ PROCEDURE insert(const source : shortstring;var s : shortstring;index : StrLenInt) ;}
{$ENDIF}

{ ANSI String procedures, have to be in one unit, to be able to overload.
  NOT in BP/TP}

{$ifndef FPC}
 {$Info EPasStr : No Ansistrings-support in compiler, no Ansistrings procedures compiled.}
{$else}
 {$Info EPasStr : Ansistring support detected; compiling AnsiStrings procedures}
PROCEDURE AppendBackSlash(VAR S : AnsiString);
PROCEDURE BinaryToStr(VAR S : AnsiString;Value,Bits   : CARDINAL);
Procedure CenterString(VAR S:AnsiString;C:Char;ScreenWidth:LONGINT);
FUNCTION  CharPos(CONST S : AnsiString;C:CHAR):LONGINT;
FUNCTION  CharPosSet(CONST S : AnsiString;CONST CSet:CHARSET):LONGINT;
PROCEDURE Commastr(var S : AnsiString;sep:CHAR);
PROCEDURE Compresstabs(Source:AnsiString;VAR Dest:AnsiString;Tabsize:LONGINT);
PROCEDURE ExpandTabs(CONST P : AnsiString;VAR P2:AnsiString;Tabsize:LONGINT);
FUNCTION  GetBetween(Source:AnsiString;VAR Dest:AnsiString;C1,C2:CHAR):BOOLEAN;
PROCEDURE HexToStr   (VAR S : AnsiString;Value,Digits : CARDINAL);
PROCEDURE Invert( VAR s :AnsiString);
PROCEDURE Item(VAR R: AnsiString; CONST S: AnsiString; CONST T: CHARSET; N: LONGINT);
PROCEDURE Item(VAR R: AnsiString; CONST S: AnsiString; T: CHAR; N: LONGINT);
PROCEDURE ItemS(VAR R: AnsiString; CONST S: AnsiString; CONST T: AnsiString; N: LONGINT);
PROCEDURE KillBChar(VAR S : AnsiString;CONST CSet:CHARSET);
PROCEDURE KillChar(VAR S : AnsiString; CONST CSet:CHARSET);
PROCEDURE KillChrTot(VAR S : AnsiString;CONST CSet:CHARSET);
FUNCTION  LeftStr( CONST StrName: AnsiString; NumChars : LONGINT) : AnsiString;
PROCEDURE LGrow(VAR S : AnsiString;Ch:CHAR;Count:LONGINT);
PROCEDURE LTrim(VAR P : AnsiString;Ch:Char);
PROCEDURE LowerCase(VAR S : AnsiString);
FUNCTION  Match(CONST source, pattern: AnsiString): Boolean;
FUNCTION  MidStr( CONST StrName: AnsiString; StartPos, NumChars : LONGINT) : AnsiString;
FUNCTION  NextCharPos(CONST S : AnsiString;C:CHAR;Count:LONGINT):LONGINT;
FUNCTION  NextCharPosSet(CONST S : AnsiString;CONST C:CHARSET;Count:LONGINT):LONGINT;
FUNCTION  NextRCharPos(CONST S : AnsiString;C:CHAR;Count:LONGINT):LONGINT;
PROCEDURE OctToStr   (VAR S : AnsiString;Value,Digits : CARDINAL);
FUNCTION  RCharPos(CONST S : AnsiString;C:CHAR):LONGINT;
FUNCTION  Replace(Sub1,Sub2: AnsiString;SS: AnsiString ): AnsiString;
FUNCTION  ReplaceLast(Sub1,Sub2: AnsiString;var SS: AnsiString): AnsiString;
FUNCTION  RightStr ( CONST StrName: AnsiString; NumChars : LONGINT) : AnsiString;
FUNCTION  RPos(CONST SubStr,S : AnsiString):LONGINT;
PROCEDURE RGrow(VAR S : AnsiString;C:CHAR;Count:LONGINT);
PROCEDURE RTrim(VAR P : AnsiString;Ch:Char);
PROCEDURE ReplaceChar(VAR S :AnsiString;ReplaceMe,ReplaceWith:CHAR);
PROCEDURE Slice (VAR R: AnsiString; S: AnsiString; P,L: LONGINT);
PROCEDURE StrStr(VAR P :AnsiString;C:Char;Count:LONGINT);
FUNCTION  StrToBinary(CONST S : AnsiString;Bits  : CARDINAL):CARDINAL;
FUNCTION  StrToHex   (CONST S : AnsiString;Digits: CARDINAL):CARDINAL;
FUNCTION  StrToOct   (CONST S : AnsiString;Digits: CARDINAL):CARDINAL;
PROCEDURE StripChar(VAR S : AnsiString;C:CHAR);
PROCEDURE StripDoubleChar(VAR S : AnsiString;C:CHAR);
PROCEDURE UpperCase(VAR S : AnsiString);
{$ENDIF}

IMPLEMENTATION

{Ansistring procedures}

{$ifdef fpc}

FUNCTION MidStr( CONST StrName: AnsiString; StartPos, NumChars : LONGINT) : AnsiString;
BEGIN
  MidStr := Copy ( StrName, StartPos, NumChars);
END; {Mid$}

FUNCTION LeftStr( CONST StrName: AnsiString; NumChars : LONGINT) : AnsiString;
BEGIN
  LeftStr := Copy ( StrName, 1, NumChars);
END; {Left$}

FUNCTION RightStr ( CONST StrName: AnsiString; NumChars : LONGINT) : AnsiString;
BEGIN
  RightStr := Copy ( StrName, ( Length(StrName) - (NumChars - 1)), NumChars);
END; {Right$}

PROCEDURE Slice (VAR R: AnsiString; S: AnsiString; P,L: LONGINT);

VAR A : LONGINT;

BEGIN
 A:=Length(S);
 IF P>A THEN
  BEGIN
   SetLength(R,0);
   EXIT;
  END;
 IF (P+L+1)>A THEN
  L:=A-P+1;
 SetLength(R,L);
 Move(S[P],R[1],L);
END;

Procedure CenterString(VAR S:AnsiString;C:Char;ScreenWidth:LONGINT);

VAR Grw1 : LONGINT;

BEGIN
 Grw1:= ScreenWidth-Length(S);
 IF Grw1>0 THEN
  BEGIN
   IF ODD(Grw1) THEN
    INC(Grw1);
   LGrow(S,C,ScreenWidth-(Grw1 shr 1)); {LGrow and RGrow are safe}
   RGrow(S,C,ScreenWidth);              { They do nothing if count<=Length(S)}
  END;
END;

FUNCTION  ReplaceLast(Sub1,Sub2: AnsiString;VAR SS: AnsiString): AnsiString;

VAR PP: LONGINT;
    L1,L2: LONGINT;

BEGIN
   PP:= RPos(Sub1,SS);

   { Seems more complicated than necessary,
     but Delete and then Insert is slower (two string-rest moves instead of one
     with this method) }

   L1:=Length(Sub1); L2:=Length(Sub2);
   IF (PP<>0) AND (L1<>0) AND (L2<>0) THEN
    BEGIN
     IF L1>L2 THEN
       Delete(SS,PP,L1-L2);                   { Sub2 < Sub1, delete chars}
     IF L1<L2 THEN
      BEGIN                                  { Sub2 > Sub1, make room for additional chars}
       SetLength(SS,Length(SS)+L2-L1);       { Force enlarge Ansistring (since I enlarge it myself)}
       Move(SS[PP],SS[PP+L2-L1],Length(SS)-PP);
      END;
     Move(Sub2[1],SS[PP],L2);                { move the string in the new string}
    END;
   ReplaceLast:= SS
end;

FUNCTION  Replace(Sub1,Sub2: AnsiString;SS: AnsiString ): AnsiString;

VAR PP: LONGINT;
    L1,L2: LONGINT;

BEGIN
   PP:= Pos(Sub1,SS);

   { Seems more complicated than necessary,
     but Delete and then Insert is slower (two string-rest moves instead of one
     with this method) }

   L1:=Length(Sub1); L2:=Length(Sub2);
   IF (PP<>0) AND (L1<>0) AND (L2<>0) THEN
    BEGIN
     IF L1>L2 THEN
       Delete(SS,PP,L1-L2);                   { Sub2 < Sub1, delete chars}
     IF L1<L2 THEN
      BEGIN                                  { Sub2 > Sub1, make room for additional chars}
       SetLength(SS,Length(SS)+L2-L1);       { Force enlarge Ansistring (since I enlarge it myself)}
       Move(SS[PP],SS[PP+L2-L1],Length(SS)-PP);
      END;
     Move(Sub2[1],SS[PP],L2);                { move the string in the new string}
    END;
   Replace:= SS
end;

PROCEDURE Invert( VAR s :AnsiString);

VAR  C          : CHAR;
     Start , En : integer;
BEGIN
        Start:=1;
        En:=Length(s);
        WHILE ( start <= en ) DO
         BEGIN
          c := s[en];
          s[en]  := s[Start];
          s[Start]:=c;
          Start:= Start+1;
          en:= en -1
        END;
END;

FUNCTION GetBetween(Source:AnsiString;VAR Dest:AnsiString;C1,C2:CHAR):BOOLEAN;

VAR C,B : CARDINAL;

BEGIN

 IF Length(Source)=0 THEN
  BEGIN
   GetBetween:=FALSE;
   EXIT;
  END;
 B:=CharPos(Source,C1);
 IF (B=(Length(Source)-1)) OR (B=0) THEN
  BEGIN
   GetBetween:=FALSE;
   EXIT;
  END;
 C:=NextCharPos(Source,C2,B+1);
 IF (C=0) OR (C=(B+1)) THEN
  BEGIN
   SetLength(Dest,0);    {Only AnsiString modification}
   GetBetween:=FALSE;
   EXIT;
  END;
 Dest:=Copy(Source,B+1,C-B-1);
 GetBetween:=TRUE;
END;{GetBetween}

PROCEDURE Item(VAR R: AnsiString; CONST S: AnsiString; T: CHAR; N:LONGINT);
{ Returns Nth substring of S in R (first is 0), separator is
character T One of my favourite string procedures.}

VAR A,B,C,D,E : LONGINT;

BEGIN
 A:=Length(S)+1; B:=1; C:=0; D:=1;   {Skip leading T's}
 WHILE (B<A) AND (S[B]=T) DO
  INC(B);
 E:=NextCharPos(S,T,B);
 IF (E=0) AND (N=0) THEN
   R:=Copy(S,B,A-B-1)
 ELSE
  BEGIN
   D:=B;
   WHILE (B<=A) AND (C<(N+1)) DO
    BEGIN
     IF S[B]=T THEN
      BEGIN
       IF S[B-1] <> T THEN
        BEGIN
         INC(C);
         E:=D;
         D:=B;
        END;
      END;
     INC(B);
    END;
   IF B>A THEN
   BEGIN
    E:=D;
    D:=B+1;
   END;
  IF N=0 THEN
    DEC(E);
   IF (C=0) OR (C<N) THEN
    SetLength(R,0)
   ELSE
    BEGIN
     INC(E);
     WHILE S[E]=T DO INC(E);
     R:=Copy(S,E,D-E);
    END;
  END;
END;

PROCEDURE Item(VAR R: AnsiString; CONST S: AnsiString; CONST T: CHARSET; N: LONGINT);
{ Returns Nth substring of S in R (first is 0), separators are all
characters in S. One of my favourite string procedures.}

VAR
  I,J,
  L : LONGINT;
  U : BOOLEAN;

BEGIN
  I := 1;
  L := Length(S)+1;
  REPEAT
   WHILE (I <= L) AND (S[I] IN T) DO INC(I); { Skip separators }
   IF (N = 0) OR (I = L) THEN   { N is number of separator-aggregates to
                                    skip, or end of string (number
                                    substring to high}
    U:=TRUE
   ELSE
    BEGIN
     U:=FALSE;
     DEC(N);                    { One Item/separator-aggregate less to go}
    WHILE (I < L) AND NOT (S[I] IN T) DO INC(I);  { Skip item }
    END;
  UNTIL U;
  IF (N=0) AND (I<L) THEN
    BEGIN
     IF I=(L-1) THEN
       SetLength(R,1)
     ELSE
      BEGIN
       J:=NextCharPosSet(S,T,I+1);      { Ansistring estimation of length}
       IF J=0 THEN
        J:=Length(S)+1;
       SetLength(R,J-I);
      END;

     J  := 0;
     WHILE (I <= (L)) AND NOT (S[I] IN T) DO
      BEGIN
       R[J+1] := S[I];
       INC(I);
       INC(J);
      END;
     IF I>L THEN
      DEC(J);
   END
  ELSE
   SetLength(R,0);
END;

PROCEDURE ItemS(VAR R: AnsiString; CONST S: AnsiString; CONST T: AnsiString; N: LONGINT);

VAR Ses : CHARSET;
    A   : LONGINT;
    L   : LONGINT;

BEGIN
 Ses:=[];
 L:=Length(T);
 IF L>0 THEN
  BEGIN
   FOR A :=1 TO L DO
    Include(Ses,T[A]);  {works with 0.99.6+}
   Item(R,S,Ses,N);
  END;
END;


PROCEDURE LTrim(VAR P : AnsiString;Ch:Char);

VAR I,J : LONGINT;

BEGIN
 I:=Length(P);      { Keeping length in local data eases optimalisations}
 IF I>0 THEN
  BEGIN
   J:=1;
   WHILE (P[J]=Ch) AND (J<=I) DO
    INC(J);
   IF J>1 THEN
    Delete(P,1,J-1);
  END;
END;

PROCEDURE RTrim(VAR P : AnsiString;Ch:Char);

VAR I,J : LONGINT;

BEGIN
 I:=Length(P);             { Keeping length in local data eases optimalisations}
 IF (I>0) THEN
  BEGIN
   J:=I;
   WHILE (P[J]=Ch) AND (J>0) DO DEC(J);
   IF J<>I THEN
    SetLength(P,J);
   END;
END;


PROCEDURE RGrow(VAR S : AnsiString;C:CHAR;Count:LONGINT);

VAR I,J :LONGINT;

BEGIN
 I:=Length(S);           { Keeping length in local data eases optimalisations}
 IF I<Count THEN
  BEGIN
   SetLength(S,Count);
   FOR J:=I+1 TO Count DO
    S[J]:=C;
  END;
END;

PROCEDURE LGrow(VAR S : AnsiString;Ch:CHAR;Count:LONGINT);

VAR  I,J :LONGINT;

BEGIN
 I:=Length(S);           { Keeping length in local data eases optimalisations}
 IF I<Count THEN
  BEGIN
   SetLength(S,Count);
   IF I<>0 THEN
    Move(S[1],S[Count-I+1],I);
   FOR J:=1 TO Count-I DO
    S[J]:=Ch;
  END;
END;

PROCEDURE KillChar(VAR S : AnsiString; CONST CSet:CHARSET);

VAR I,J : LONGINT;

BEGIN
 I:=Length(S);      { Keeping length in local data eases optimalisations}
 IF (I>0) THEN
  BEGIN
   J:=1;
   WHILE (S[J] IN CSet) AND (J<=I) DO INC(J);
   IF J>1 THEN
    Delete(S,1,J-1);
   END;
END;

PROCEDURE KillBChar(VAR S : AnsiString;CONST CSet:CHARSET);

VAR I,J: LONGINT;

BEGIN
 I:=Length(S);      { Keeping length in local data eases optimalisations}
 IF (I>0) THEN
  BEGIN
   J:=I;
   WHILE (S[J] IN CSet) AND (J>0) DO DEC(J);
   IF J<>I THEN
    SetLength(S,J);
  END;
END;

FUNCTION NextCharPosSet(CONST S : AnsiString;CONST C:CHARSET;Count:LONGINT):LONGINT;

VAR I,J:LONGINT;

BEGIN
 IF PCHAR(S)=NIL THEN
  J:=0
 ELSE
  BEGIN
   I:=Length(S);
   J:=Count;
   IF J>I THEN
    BEGIN
     NextCharPosSet:=0;
     EXIT;
    END;
   WHILE (j<=i) AND (NOT (S[J] IN C)) DO INC(J);
   IF (J>I) THEN
    J:=0;                                         // NOT found.
  END;
 NextCharPosSet:=J;
END;

{Like Pos(). Returns 0 when substr not found, otherwise standard STRING index
(1..Length(X)) to first char of last substr found in S)}
FUNCTION RPos(CONST SubStr,S : AnsiString):LONGINT;

VAR I,J   : LONGINT;
    Ii,Ji : LONGINT;
    e     : BOOLEAN;

BEGIN
 J:=Length(SubStr);
 IF J=1 THEN
  J:=RCharPos(S,SubStr[1]) {XTDFPC procedure, call Char only variant, uses asm
                             scasb}
 ELSE
  IF J>1 THEN
   BEGIN
    I:=Length(S);
    e:=TRUE;
    WHILE e AND (I>0) DO
     BEGIN
      IF (SubStr[J]=S[I]) THEN   {Scan for last character of substring}
       BEGIN
        IF J<=I THEN {Haven't reached begin of string yet?}
         BEGIN
          e:=FALSE;          {Found, scan for other characters,
                             kind of pascalled  std; rep cmpsb}
          Ii:=I;
          Ji:=J;
          WHILE (Ji>1) AND NOT e DO
           BEGIN
            Dec(Ii);
            Dec(Ji);
            IF S[Ii]<>SubStr[Ji] THEN
             e:=TRUE;
           END; {While}
        END;
       END; {IF}
     DEC(I);
    END; {WHILE}
    J:=0;
    IF (NOT e) THEN
     J:=I+2-Length(SubStr); { Or  J:=I+1;  to create index to end of substr}
   END;
 RPos:=J;
END;

PROCEDURE ReplaceChar(VAR S :AnsiString;ReplaceMe,ReplaceWith:CHAR);

VAR I,J : LONGINT;

BEGIN
 I:=Length(S);
 IF I>0 THEN
  FOR J:=1 TO I DO
   IF S[J]=ReplaceMe THEN
    S[J]:=ReplaceWith;
END;

PROCEDURE StrStr(VAR P :AnsiString;C:Char;Count:LONGINT);

BEGIN
 IF Count>0 THEN
  BEGIN
   SetLength(P,Count);
   FillChar(P[1],Count,C);
  END;
END;

PROCEDURE StripChar(VAR S : AnsiString;C:CHAR);

VAR I,J,L : LONGINT;

BEGIN
 L:=Length(S);
 IF L>0 THEN
  BEGIN
   J:=1;
   FOR I:=1 TO L DO
    IF S[I]<>C THEN
     BEGIN
      S[J]:=S[I];
      INC(J);
     END;
   SetLength(S,J-1);
  END;
END;

PROCEDURE KillChrTot(VAR S : AnsiString;CONST CSet:CHARSET);

VAR I,J,L : LONGINT;

BEGIN
 L:=Length(S);
 IF L>0 THEN
  BEGIN
   J:=1;
   FOR I:=1 TO L DO
    IF NOT (S[I] IN CSet) THEN
     BEGIN
      S[J]:=S[I];
      INC(J);
     END;
   SetLength(S,J-1);
  END;
END;

PROCEDURE AppendBackSlash(VAR S :AnsiString);

VAR I : LONGINT;

BEGIN
 I:=Length(S);
 IF I>0 THEN
  BEGIN
 {$ifdef linux}
 IF S[I]<>'/' THEN
  BEGIN
   SetLength(S,I+1);
   S[I+1]:='/';
  END;
{$ELSE}
 IF S[I]<>'\' THEN
  BEGIN
   SetLength(S,I+1);
   S[I+1]:='\';
  END
{$ENDIF}
  END
 ELSE
  BEGIN
   SetLength(S,1);
   {$ifdef linux}
    S[1]:='/';
   {$ELSE}
    S[1]:='\';
   {$ENDIF}
  END;
END;

PROCEDURE StripDoubleChar(VAR S : AnsiString;C:CHAR);

VAR I,J,L : LONGINT;

BEGIN
 L:=Length(S);
 IF L>1 THEN  {No double chars to strip when Length(S)=1}
  BEGIN
   J:=2;
   FOR I:=2 TO L DO
    IF NOT((S[I]=C) AND (S[I]=S[I-1])) THEN
     BEGIN
      S[J]:=S[I];
      INC(J);
     END;
   SetLength(S,J-1);
  END;
END;

PROCEDURE UpperCase(VAR S : AnsiString);

VAR L,I : LONGINT;

BEGIN
 L:=Length(S);
 IF L>0 THEN
  FOR I:=1 TO L DO
   IF (S[I]>CHR(96)) AND (S[I]<CHR(123)) THEN
    S[I]:=CHR(ORD(S[I])-32);
END;

PROCEDURE LowerCase(VAR S : AnsiString);

VAR L,I : LONGINT;

BEGIN
 L:=Length(S);
 IF L>0 THEN
  FOR I:=1 TO L DO
   IF (S[I]>CHR(64)) AND (S[I]<CHR(91)) THEN
    S[I]:=CHR(ORD(S[I])+32);
END;

PROCEDURE Commastr(var S : AnsiString;sep:CHAR);

VAR L,I,J : LONGINT;
    SepStr: AnsiString;         {Have to have an ansistring to insert}

BEGIN
 L:=Length(S);
 I:=L DIV 3;
 IF ((L MOD 3)=0) AND (I>0) THEN
  DEC(I);
 IF I>0 THEN
  BEGIN
   SetLength(SepStr,1);
   SepStr[1]:=Sep;
   FOR J:= 1 TO I DO
    Insert(SepStr,S,L-J*3+1);
  END;
END;


PROCEDURE BinaryToStr(VAR S : AnsiString;Value,Bits : CARDINAL);

VAR I : LONGINT;

BEGIN
 SetLength(S,Bits);
 IF Bits >0 THEN
  BEGIN
   IF Bits<>32 THEN
    Value:=Value SHL (32-bits);
   FOR I:=1 TO Bits DO
    BEGIN
     IF (Value AND (1 SHL 31))<>0 THEN
      S[I]:='1'
     ELSE
      S[I]:='0';
    Value:=Value shl 1;
    END;
   END;
END;

FUNCTION StrToBinary(CONST S : AnsiString;Bits : CARDINAL):CARDINAL;

VAR I : CARDINAL;
    A : CARDINAL;

BEGIN
 IF (PChar(S)<>NIL) OR (Bits=0) THEN
  I:=0
 ELSE
  BEGIN
   I:=0;
   IF Bits>0 THEN
     FOR A :=1 TO Bits DO
      BEGIN
       I:=I SHL 1;
       I:=I OR (ORD(S[A]) AND 1);
      END;
  END;
 StrToBinary:=I;
END;

PROCEDURE HexToStr(VAR S : AnsiString;Value,Digits : CARDINAL);

VAR I : LONGINT;
    O : CARDINAL;

BEGIN
 SetLength(S,Digits);
 IF Digits >0 THEN
  BEGIN
   IF Digits<>32 THEN
    Value:=Value SHL (32-(Digits shl 2));
   FOR I:=1 TO Digits DO
    BEGIN
     O:=(Value SHR 28) +48;
     IF O>57 THEN
      INC(O,7);
     S[I]:=CHR(O);
    Value:=Value shl 4;
    END;
   END;
END;

PROCEDURE OctToStr(VAR S : AnsiString;Value,Digits : CARDINAL);

VAR I : LONGINT;
    O : CARDINAL;

BEGIN
 SetLength(S,Digits);
 IF Digits >0 THEN
  BEGIN
   IF Digits<>32 THEN
    Value:=Value SHL (32-(Digits shl 1 )-Digits);
   FOR I:=1 TO Digits DO
    BEGIN
     O:=(Value SHR 29) +48;
     S[I]:=CHR(O);
    Value:=Value shl 3;
    END;
   END;
END;


FUNCTION StrToOct(CONST S : AnsiString;Digits: CARDINAL):CARDINAL;

VAR I : CARDINAL;
    A : CARDINAL;

BEGIN
 IF (PChar(S)<>NIL) OR (Digits=0) THEN
  I:=0
 ELSE
  BEGIN
   I:=0;
   IF Digits>0 THEN
    BEGIN
     FOR A :=1 TO Digits DO
       I:=(I SHL 3) + ((ORD(S[A])-48) AND 7)
    END;
  END;
 StrToOct:=I;
END;

FUNCTION StrToHex(CONST S : AnsiString;Digits: CARDINAL):CARDINAL;

VAR I : CARDINAL;
    A : CARDINAL;

BEGIN
 IF (PChar(S)<>NIL) OR (Digits=0) THEN
  I:=0
 ELSE
  BEGIN
   I:=0;
   IF Digits>0 THEN
    BEGIN
     FOR A :=1 TO Digits DO
      BEGIN
       I:=I SHL 4;
       IF S[A]<CHR(58) THEN
        I:=I + ((ORD(S[A])-48) AND 15)
       ELSE
        I:=I + ((ORD(S[A])-55) AND 15);
      END;
    END;
  END;
 StrToHex:=I;
END;

PROCEDURE ExpandTabs(CONST P : AnsiString;VAR P2:AnsiString;Tabsize:LONGINT);

VAR I,B,C,D : LONGINT;

BEGIN
 IF PCHAR(P)=NIL THEN
  SetLength(P2,0)
 ELSE
  BEGIN
   I:=CharPos(P,#9);
   IF I=0 THEN
    BEGIN
     P2:=P;
     EXIT
    END;
   C:=0;                         { C =inserted spaces}
   REPEAT
    D:=I+C;                      { character position corrected for inserted spaces}
    C:=C+(Tabsize-((D-1) MOD TabSize))-1;         {Minus one for the hardtab}
    I:=NextCharPos(P,#9,I+1);
   UNTIL I=0;
   WritelN(Length(p)+C);

   SetLength(P2,Length(P)+C);      {Mimimal Length written, pre allocate}
   I:=1; B:=1;
   WHILE I <=Length(P) DO
    BEGIN
     IF P[I]=CHR(9) THEN
      BEGIN
       C:=Tabsize-((B-1) MOD TabSize);
       FOR D:=0 TO C-1 DO
        P2[B+D]:=' ';
       B:=B+C-1;
      END
     ELSE
      P2[B]:=P[I];
     INC(B); INC(I);
    END;
  END;
END;

PROCEDURE Compresstabs(Source:AnsiString;VAR Dest:AnsiString;Tabsize:LONGINT);

VAR I,B,C,D,RestSpac,L,L3,
    L1,L2 : LONGINT;

BEGIN
 I:=1; B:=1; L:=Length(Source);
 WHILE I<=L DO
  BEGIN
   D:=0;
   WHILE (Source[I+D]=CHR(32)) AND ((D+I)<L) DO { Count spaces}
    INC(D);
   IF D<>0 THEN
    BEGIN
     SetLength (Dest,L);                {Dest is never bigger than source}
     L1:=(I) DIV Tabsize;
     L3:=(I) MOD Tabsize;
     L2:=(I+D) DIV Tabsize;
     RestSpac:=(I+D-1) MOD Tabsize;   {Insert rest-spaces}
     IF L1<L2 THEN               {Tab border crossed?}
      BEGIN
       FOR C:= L1 TO L2-1 DO     {Insert all tabs needed}
        BEGIN
         Dest[B]:=CHR(9);
         INC(B);
        END;
       IF RestSpac>0 THEN
        BEGIN
         FillChar(Dest[B],RestSpac,32);
         INC(B,RestSpac);
        END;
       INC(I,D);                       {Advance over spaces}
      END
     ELSE
      BEGIN
        Move(Source[I],Dest[B],D);
       INC(B,D); INC(I,D);
      END;
     END
     ELSE
      BEGIN
       Dest[B]:=Source[I];
       INC(B); INC(I);
      END;
  END;
 SetLength(Dest,B-1);
END;

function Match(CONST source, pattern: AnsiString): Boolean;
var
  Lenelement,
  LenPattern: LONGINT;

  function MatchPattern(elementpos, patternpos: LONGINT): Boolean;

  begin
    if (Pattern[patternpos]='*') AND (Patternpos=lenPattern) then
     MatchPattern:=TRUE
    else if (elementpos= lenelement) and (patternpos<> lenpattern) then
     MatchPattern:=FALSE
    else if elementpos = lenelement then
     MatchPattern:=TRUE
    else begin
      case pattern[patternpos] of
      '*': if MatchPattern(elementpos,patternpos+1) then
             MatchPattern:= True
           else
             MatchPattern:= MatchPattern(elementpos+1,patternpos);
      '?': MatchPattern:= MatchPattern(elementpos+1,patternpos+1);
      else
        if source[elementpos] = pattern[patternpos] then
          MatchPattern:= MatchPattern(elementpos+1,patternpos+1)
        else
          MatchPattern:= False;
      end;
    end;
  end;

begin
  Lenelement:=Length(Source)+1;
  LenPattern:=Length(Pattern)+1;
  IF (Lenelement>1) AND (LenPattern>1) THEN
   Match:= MatchPattern(1,1)
  ELSE
   Match:=FALSE;
end;


{---------------------------------------------------------------------
                          Ansistring procedures with asm equivalent}

{$IFNDEF ANSIUseASM}
{$INFO EPasStr : No assembler for AnsiStrings}

FUNCTION CharPos(CONST S : AnsiString;C:CHAR):LONGINT;

VAR I,J:LONGINT;

BEGIN
 IF PCHAR(S)=NIL THEN
  J:=0
 ELSE
  BEGIN
   I:=Length(S);
   J:=1;
   WHILE (S[J]<>C) AND (J<=I) DO INC(J);
   IF (J>I) THEN
    J:=0;
  END;
 CharPos:=J;
END;

FUNCTION NextCharPos(CONST S : AnsiString;C:CHAR;Count:LONGINT):LONGINT;

VAR I,J:LONGINT;

BEGIN
 IF PCHAR(S)=NIL THEN
  J:=0
 ELSE
  BEGIN
   I:=Length(S);
   J:=Count;
   IF J>I THEN
    BEGIN
     J:=0;
     EXIT;
    END;
   WHILE (j<=i) AND (S[J]<>C) DO INC(J);
   IF (J>I) THEN
    J:=0;                                         // NOT found.
  END;
 NextCharPos:=J;
END;

FUNCTION CharPosSet(CONST S : AnsiString;CONST CSet:CHARSET):LONGINT;

VAR I,J:LONGINT;

BEGIN
 I:=Length(S);
 J:=1;
 IF I=0 THEN
  J:=0
 ELSE
  BEGIN
   WHILE  NOT (S[J] IN CSet) AND (J<=I) DO
    INC(J);
   IF (J>I) THEN
    J:=0;
  END;
 CharPosSet:=J;
END;

FUNCTION RCharPos(CONST S : AnsiString;C:CHAR):LONGINT;

VAR I,J:LONGINT;

BEGIN
 I:=Length(S);
 J:=I;
 IF I<>0 THEN
  WHILE (S[J]<>C) AND (J>0) DO
   DEC(J);
 RCharPos:=J;
END;

FUNCTION NextRCharPos(CONST S : AnsiString;C:CHAR;Count:LONGINT):LONGINT;

VAR I,J:LONGINT;

BEGIN
 I:=Length(S);
 J:=Count;
 IF J>I THEN J:=I;
 IF I<>0 THEN
  WHILE (S[J]<>C) AND (J>0) DO
   DEC(J);
 NextRCharPos:=J;
END;


{$ELSE}
{$INFO EPasStr : Assembler for some AnsiStrings routines}
{$I epas_aa.inc}

{$ENDIF}


{$ENDIF}
{---------------------------------------------------------------------
                          Pascalstring procedures without asm equivalent}

FUNCTION MidStr( CONST StrName: String; StartPos, NumChars : Integer) : String;

BEGIN
  MidStr := Copy ( StrName, StartPos, NumChars);
END; {Mid$}

FUNCTION LeftStr( CONST StrName: String; NumChars : Integer) : String;
BEGIN
  LeftStr := Copy ( StrName, 1, NumChars);
END; {Left$}

FUNCTION RightStr ( CONST StrName: String; NumChars : Integer) : String;
BEGIN
  RightStr := Copy ( StrName, ( Length(StrName) - (NumChars - 1)), NumChars);
END; {Right$}

PROCEDURE Slice (VAR R: String; S: String; P,L: LONGINT);

VAR A : LONGINT;

BEGIN
 A:=Length(S);
 IF P>A THEN
  BEGIN
   R[0]:=#0;
   EXIT;
  END;
 IF (P+L+1)>A THEN
  L:=A-P+1;
 Move(S[P],R[1],L);
 R[0]:=CHR(L);
END;

Procedure CenterString(VAR S:String;C:Char;ScreenWidth:LONGINT);

VAR Grw1 : LONGINT;

BEGIN
 Grw1:= ScreenWidth-Length(S);
 IF Grw1>0 THEN
  BEGIN
   IF ODD(Grw1) THEN
    INC(Grw1);
   LGrow(S,C,ScreenWidth-(Grw1 shr 1)); {LGrow and RGrow are safe}
   RGrow(S,C,ScreenWidth);              { They do nothing if count<=Length(S)}
  END;
END;

FUNCTION NextCharPosSet(CONST S : String;CONST C:CHARSET;Count:LONGINT):LONGINT;

VAR I,J:LONGINT;

BEGIN
   I:=Length(S);
   IF I=0 THEN
    J:=0
   ELSE
    BEGIN
   J:=Count;
   IF J>I THEN
    BEGIN
     NextCharPosSet:=0;
     EXIT;
    END;
   WHILE (j<=i) AND (NOT (S[J] IN C)) DO INC(J);
   IF (J>I) THEN
    J:=0;                                         // NOT found.
   END;
 NextCharPosSet:=J;
END;


PROCEDURE Compresstabs(Source:String;VAR Dest:String;Tabsize:LONGINT);

VAR I,B,C,D,RestSpac,L,L3,
    L1,L2 : LONGINT;

BEGIN
 I:=1; B:=1; L:=Length(Source);
 WHILE I<=L DO
  BEGIN
   D:=0;
   WHILE (Source[I+D]=CHR(32)) AND ((D+I)<L) DO { Count spaces}
    INC(D);
   IF D<>0 THEN
    BEGIN
     L1:=(I) DIV Tabsize;
     L3:=(I) MOD Tabsize;
     L2:=(I+D) DIV Tabsize;
     RestSpac:=(I+D-1) MOD Tabsize;   {Insert rest-spaces}
     IF L1<L2 THEN               {Tab border crossed?}
      BEGIN
       FOR C:= L1 TO L2-1 DO     {Insert all tabs needed}
        BEGIN
         Dest[B]:=CHR(9);
         INC(B);
        END;
       IF RestSpac>0 THEN
        BEGIN
         FillChar(Dest[B],RestSpac,32);
         INC(B,RestSpac);
        END;
       INC(I,D);                       {Advance over spaces}
      END
     ELSE
      BEGIN
        Move(Source[I],Dest[B],D);
       INC(B,D); INC(I,D);
      END;
     END
     ELSE
      BEGIN
       Dest[B]:=Source[I];
       INC(B); INC(I);
      END;
  END;
 Dest[0]:=CHR(B-1);
END;

FUNCTION  ReplaceLast(Sub1,Sub2: STRING;VAR SS: STRING ): STRING;

VAR PP: LONGINT;
    L1,L2: LONGINT;

BEGIN
   PP:= RPos(Sub1,SS);

   { Seems more complicated than necessary,
     but Delete and then Insert is slower (two string-rest moves instead of one
     with this method) }

   L1:=Length(Sub1); L2:=Length(Sub2);
   IF (PP<>0) AND (L1<>0) AND (L2<>0) THEN
    BEGIN
     IF L1>L2 THEN
       Delete(SS,PP,L1-L2);                   { Sub2 < Sub1, delete chars}
     IF L1<L2 THEN
      BEGIN                                  { Sub2 > Sub1, make room for additional chars}
       Move(SS[PP],SS[PP+L2-L1],Length(SS)-PP);
      END;
     Move(Sub2[1],SS[PP],L2);                { move the string in the new string}
    END;
   ReplaceLast:= SS
end;

FUNCTION  Replace(Sub1,Sub2: STRING;SS: STRING ): STRING;

VAR PP: LONGINT;
    L1,L2: LONGINT;

BEGIN
   PP:= Pos(Sub1,SS);

   { Seems more complicated than necessary,
     but Delete and then Insert is slower (two string-rest moves instead of one
     with this method) }

   L1:=Length(Sub1); L2:=Length(Sub2);
   IF (PP<>0) AND (L1<>0) AND (L2<>0) THEN
    BEGIN
     IF L1>L2 THEN
       Delete(SS,PP,L1-L2);                   { Sub2 < Sub1, delete chars}
     IF L1<L2 THEN
      BEGIN                                  { Sub2 > Sub1, make room for additional chars}
       SetLength(SS,Length(SS)+L2-L1);       { Force enlarge Ansistring (since I enlarge it myself)}
       Move(SS[PP],SS[PP+L2-L1],Length(SS)-PP);
      END;
     Move(Sub2[1],SS[PP],L2);                { move the string in the new string}
    END;
   Replace:= SS
end;

FUNCTION GetBetween(Source:String;VAR Dest:String;C1,C2:CHAR):BOOLEAN;

VAR C,B : CARDINAL;

BEGIN
 Dest[0]:=CHR(0);
 IF Length(Source)=0 THEN
  BEGIN
   GetBetween:=FALSE;
   EXIT;
  END;
 B:=CharPos(Source,C1);
 IF (B=(Length(Source)-1)) OR (B=0) THEN
  BEGIN
   GetBetween:=FALSE;
   EXIT;
  END;
 C:=NextCharPos(Source,C2,B+1);
 IF (C=0) OR (C=(B+1)) THEN
  BEGIN
   GetBetween:=FALSE;
   EXIT;
  END;
 Dest:=Copy(Source,B+1,C-B-1);
 GetBetween:=TRUE;
END;{GetBetween}


PROCEDURE Item(VAR R: STRING; CONST S: STRING; CONST T: CHARSET; N: LONGINT);
{ Returns Nth substring of S in R (first is 0), separators are all
characters in S. One of my favourite string procedures.}

VAR
  I,J,
  HR,L : LONGINT;
  U : BOOLEAN;

BEGIN
  I := 1;
  L := Length(S)+1;
  REPEAT
   WHILE (I <= L) AND (S[I] IN T) DO INC(I); { Skip separators }
   IF (N = 0) OR (I = L) THEN   { N is number of separator-aggregates to
                                    skip, or end of string (number
                                    substring to high}
    U:=TRUE
   ELSE
    BEGIN
     U:=FALSE;
     DEC(N);                    { One Item/separator-aggregate less to go}
    WHILE (I < L) AND NOT (S[I] IN T) DO INC(I);  { Skip item }
    END;
  UNTIL U;
  R[0]:=CHR(0);
  IF (N=0) THEN
   IF I<L THEN
  BEGIN
    J  := 0;
  HR := HIGH(R);
  WHILE (I <= (L)) AND NOT (S[I] IN T) AND (J <= HR) DO
  BEGIN
    R[J+1] := S[I];
    INC(I);
    INC(J);
  END;
  IF I>L THEN
   DEC(J);
  R[0]:=CHR(J);
  END
END;

{ Returns Nth substring of S in R (first is 0), separator are characters in
    string T}
PROCEDURE ItemS(VAR R: STRING; CONST S: STRING; CONST T: String; N: LONGINT);

VAR Ses : CHARSET;
    A   : LONGINT;

BEGIN
 Ses:=[];
 IF ORD(T[0])>0 THEN
  BEGIN
   FOR A :=1 TO ORD(T[0]) DO
    {Ses:=Ses+[T[A]];}
   Include(Ses,T[A]);  {works with 0.99.6+}
   Item(R,S,Ses,N);
  END;
END;

{OVERLOADED}
{$IFDEF Turbo}
PROCEDURE ItemC(VAR R: STRING; CONST S: STRING; T: CHAR; N: LONGINT);
{$ELSE}
PROCEDURE Item(VAR R: STRING; CONST S: STRING; T: CHAR; N: LONGINT);
{$ENDIF}
{ Returns Nth substring of S in R (first is 0), separator is character T
One of my favourite string procedures.}

VAR A,B,C,D,E : LONGINT;

BEGIN
 A:=ORD(S[0])+1; B:=1; C:=0; D:=1;   {Skip leading T's}
 WHILE (B<=A) AND (S[B]=T) DO
  INC(B);
 IF (NextCharPos(S,T,B)=0) AND (N=0) THEN
  R:=Copy(S,B,A-B)
 ELSE
  BEGIN
   D:=B;
   WHILE (B<=A) AND (C<(N+1)) DO
    BEGIN
     IF S[B]=T THEN
      BEGIN
       IF S[B-1] <> T THEN
        BEGIN
         INC(C);
         E:=D;
         D:=B;
        END;
      END;
     INC(B);
    END;
   IF B>A THEN
   BEGIN
    E:=D;
    D:=B+1;
   END;
  IF N=0 THEN
    DEC(E);
   IF (C=0) OR (C<N) THEN
    R[0]:=CHR(0)
   ELSE
    BEGIN
     INC(E);
     WHILE S[E]=T DO INC(E);
     R:=Copy(S,E,D-E);
    END;
  END;
END;

function Match(CONST source, pattern: String): Boolean;
var
  Lenelement,
  LenPattern: LONGINT;

  function MatchPattern(elementpos, patternpos: LONGINT): Boolean;

  begin
    if (Pattern[patternpos]='*') AND (Patternpos=lenPattern) then
     MatchPattern:=TRUE
    else if (elementpos= lenelement) and (patternpos<> lenpattern) then
     MatchPattern:=FALSE
    else if elementpos = lenelement then
     MatchPattern:=TRUE
    else begin
      case pattern[patternpos] of
      '*': if MatchPattern(elementpos,patternpos+1) then
             MatchPattern:= True
           else
             MatchPattern:= MatchPattern(elementpos+1,patternpos);
      '?': MatchPattern:= MatchPattern(elementpos+1,patternpos+1);
      else
        if source[elementpos] = pattern[patternpos] then
          MatchPattern:= MatchPattern(elementpos+1,patternpos+1)
        else
          MatchPattern:= False;
      end;
    end;
  end;

begin
  Lenelement:=Length(Source)+1;
  LenPattern:=Length(Pattern);
  IF (Lenelement>1) AND (LenPattern>1) THEN
   Match:= MatchPattern(1,1)
  ELSE
   Match:=FALSE;
end;


{Insert for a character.
  Should be an Insert overloaded procedure. I can't since Insert is in a
    different unit, asm version not ready yet}
PROCEDURE InsertC(VAR S : String;C:CHAR;Position:LONGINT);

VAR L : String[5];

BEGIN
 L[0]:=#1; L[1]:=C;
 Insert(L,S,Position);
END;

FUNCTION QItem(VAR R: STRING; CONST S: STRING; T: CHARSET; N: LONGINT;Mode : LONGINT):LONGINT;

CONST QuoteSet: CHARSET=['"',#39];

VAR   QuoteChar : CHAR;
      L,HR     : LONGINT;

FUNCTION ReadItem(VAR StartPos : LONGINT;Cpy:BOOLEAN):LONGINT;
{Reads a non separator block, the "item" on which all item types are based
if Cpy then the block is copied to R too.}

VAR Quoted   : BOOLEAN;
    OPos     : LONGINT;

BEGIN
 Quoted:=FALSE;
 IF ((Mode=2) AND (S[StartPos] IN QuoteSet)) OR ((Mode<2) AND (S[StartPos]=QuoteChar)) THEN
  Quoted:=TRUE;
 IF Quoted THEN
  BEGIN
   IF Mode<2 THEN
    BEGIN
     OPos:=StartPos+1;
     StartPos:=NextCharPos(S,QuoteChar,OPos);
     IF StartPos=0 THEN {Unmatched quote, invalid input}
      BEGIN
       ReadItem:=-1;
       StartPos:=OPos;
       EXIT
      END;
     IF Cpy THEN
      R:=Copy(S,OPos,StartPos-OPos);
    END
   else
    BEGIN
     OPos:=StartPos+1;
     StartPos:=NextCharPos(S,S[StartPos],OPos);
     IF StartPos=0 THEN {Unmatched quote, invalid input}
      BEGIN
       ReadItem:=-1;
       StartPos:=OPos;
       EXIT
      END;
     IF Cpy THEN
      R:=Copy(S,OPos,StartPos-OPos);
    END;
   INC(StartPos);
   END
  ELSE {Not quoted, old skipper}
   IF Cpy THEN
    BEGIN
     OPos:=0;
     WHILE (StartPos <= (L)) AND NOT (S[StartPos] IN T) AND (OPos <= HR) DO
      BEGIN
       R[OPos+1] := S[StartPos];
       INC(StartPos);
       INC(OPos);
      END;
     IF StartPos>L THEN
      DEC(OPos);
     R[0]:=CHR(OPos);
    END
   ELSE
    WHILE (StartPos < L) AND NOT (S[StartPos] IN T) DO INC(StartPos);  { Skip item }
  ReadItem:=0;
END;

VAR
  I,J  : LONGINT;
  U : BOOLEAN;

BEGIN
  CASE Mode OF {Makes it a lot safer}
   0 : BEGIN
        EXCLUDE(T,'"');
        QuoteChar:='"';
       END;
   1 : BEGIN
        EXCLUDE(T,#39);
        QuoteChar:=#39;
       END;
   2 : T:=T-QuoteSet;
  ELSE
   BEGIN    {Mode wrong, exit empty}
    R[0]:=#0;
    QItem:=-1;
    EXIT
   END;
  END;
  I := 1;
  L := Length(S)+1;
  HR := HIGH(R);

  REPEAT
   WHILE (I <= L) AND (S[I] IN T) DO INC(I); { Skip separators }
   IF (N = 0) OR (I = L) THEN   { N is number of separator-aggregates to
                                    skip, or end of string (number
                                    substring to high}
    U:=TRUE
   ELSE
    BEGIN
     U:=FALSE;
     DEC(N);                    { One Item/separator-aggregate less to go}
    IF ReadItem(I,FALSE)=-1 THEN
     BEGIN
      QItem:=I;
      EXIT;
     END;
    END;
  UNTIL U;
  R[0]:=#0;
  IF (N=0) THEN
   IF I<L THEN
  BEGIN
  IF ReadItem(I,TRUE)=-1 THEN
   BEGIN
    QItem:=I;
    EXIT;
   END;
  END;
 QItem:=0;
END;


{---------------------------------------------------------------------
    Pascalstring procedures with asm equivalent further below}

{$IFNDEF Useasm}

{Like Pos(). Returns 0 when substr not found, otherwise standard STRING index
(1..Length(X)) to first char of last substr found in S)}
FUNCTION RPos(CONST SubStr,S : String):LONGINT;

VAR I,J   : LONGINT;
    Ii,Ji : LONGINT;
    e     : BOOLEAN;

BEGIN
 J:=Length(SubStr);
 IF J=1 THEN
  J:=RCharPos(S,SubStr[1]) {XTDFPC procedure, call Char only variant, uses asm
                             scasb}
 ELSE
  IF J>1 THEN
   BEGIN
    I:=Length(S);
    e:=TRUE;
    WHILE e AND (I>0) DO
     BEGIN
      IF (SubStr[J]=S[I]) THEN   {Scan for last character of substring}
       BEGIN
        IF J<=I THEN {Haven't reached begin of string yet?}
         BEGIN
          e:=FALSE;          {Found, scan for other characters,
                             kind of pascalled  std; rep cmpsb}
          Ii:=I;
          Ji:=J;
          WHILE (Ji>1) AND NOT e DO
           BEGIN
            Dec(Ii);
            Dec(Ji);
            IF S[Ii]<>SubStr[Ji] THEN
             e:=TRUE;
           END; {While}
        END;
       END; {IF}
     DEC(I);
    END; {WHILE}
    J:=0;
    IF (NOT e) THEN
     J:=I+2-Length(SubStr); { Or  J:=I+1;  to create index to end of substr}
   END;
 RPos:=J;
END;

PROCEDURE LTrim(VAR P : String;Ch:Char);

VAR I,J : LONGINT;

BEGIN
 I:=ORD(P[0]);      { Keeping length in local data eases optimalisations}
 IF (I>0) THEN
  BEGIN
   J:=1;
   WHILE (P[J]=Ch) AND (J<=I) DO INC(J);
   IF J>1 THEN
    Delete(P,1,J-1);
   END;
END;

PROCEDURE RTrim(VAR P : String;Ch:Char);

VAR I,J : LONGINT;

BEGIN
 I:=ORD(P[0]);      { Keeping length in local data eases optimalisations}
 IF (I>0) THEN
  BEGIN
   J:=I;
   WHILE (P[J]=Ch) AND (J>0) DO DEC(J);
   IF J<>I THEN
    Delete(P,J+1,I-J+1);
   END;
END;

PROCEDURE RGrow(VAR S : String;C:CHAR;Count:LONGINT);

VAR I,J :LONGINT;

BEGIN
 I:=ORD(S[0]);           { Keeping length in local data eases optimalisations}
 IF I<Count THEN
  BEGIN
   FOR J:=I+1 TO Count DO
    S[J]:=C;
   S[0]:=CHR(Count);
  END;
END;

PROCEDURE LGrow(VAR S : String;Ch:CHAR;Count:LONGINT);

VAR  I,J :LONGINT;

BEGIN
 I:=ORD(S[0]);           { Keeping length in local data eases optimalisations}
 IF I<Count THEN
  BEGIN
   IF I<>0 THEN
    Move(S[1],S[Count-I+1],I);
   FOR J:=1 TO Count-I DO
    S[J]:=Ch;
   S[0]:=CHR(Count);
  END;
END;

PROCEDURE KillChar(VAR S : String; CONST CSet:CHARSET);

VAR I,J : LONGINT;

BEGIN
 I:=ORD(S[0]);      { Keeping length in local data eases optimalisations}
 IF (I>0) THEN
  BEGIN
   J:=1;
   WHILE (S[J] IN CSet) AND (J<=I) DO INC(J);
   IF J>1 THEN
    Delete(S,1,J-1);
   END;
END;

PROCEDURE KillBChar(VAR S : String;CONST CSet:CHARSET);

VAR I,J : LONGINT;

BEGIN
 I:=ORD(S[0]);      { Keeping length in local data eases optimalisations}
 IF (I>0) THEN
  BEGIN
   J:=I;
   WHILE (S[J] IN CSet) AND (J>0) DO DEC(J);
   IF J<>I THEN
    Delete(S,J+1,I-J+1);
   END;
END;

FUNCTION CharPos(CONST S : String;C:CHAR):LONGINT;

VAR I,J:LONGINT;

BEGIN
 I:=ORD(S[0]);
 J:=1;
 WHILE (S[J]<>C) AND (J<=I) DO INC(J);
 IF (J>I) THEN
  J:=0;
 CharPos:=J;
END;

FUNCTION NextCharPos(CONST S : String;C:CHAR;Count:LONGINT):LONGINT;

VAR I,J:LONGINT;

BEGIN
 I:=ORD(S[0]);
 J:=Count;
 IF J>I THEN
  J:=I;
 WHILE (S[J]<>C) AND (J<=I) DO INC(J);
 IF (J>I) THEN
  J:=0;
 NextCharPos:=J;
END;

FUNCTION RCharPos(CONST S : String;C:CHAR):LONGINT;

VAR I,J:LONGINT;

BEGIN
 I:=ORD(S[0]);
 J:=I;
 WHILE (S[J]<>C) AND (J>0) DO DEC(J);
 RCharPos:=J;
END;

FUNCTION NextRCharPos(CONST S : String;C:CHAR;Count:LONGINT):LONGINT;

VAR I,J:LONGINT;

BEGIN
 I:=ORD(S[0]);
 J:=Count;
 IF J>I THEN J:=I;
 WHILE (S[J]<>C) AND (J>0) DO DEC(J);
 NextRCharPos:=J;
END;

FUNCTION CharPosSet(CONST S : String;CONST CSet:CHARSET):LONGINT;

VAR I,J:LONGINT;

BEGIN
 I:=ORD(S[0]);
 J:=1;
 WHILE  NOT (S[J] IN CSet) AND (J<=I) DO INC(J);
 IF (J>I) THEN
  J:=0;
 CharPosSet:=J;
END;

PROCEDURE ReplaceChar(VAR S :String;ReplaceMe,ReplaceWith:CHAR);

VAR I,J : LONGINT;

BEGIN
 I:=ORD(S[0]);
 IF I>0 THEN
  FOR J:=1 TO I DO IF S[J]=ReplaceMe THEN S[J]:=ReplaceWith;
END;

PROCEDURE StrStr(VAR P :String;C:Char;Count:LONGINT);

VAR I :LONGINT;

BEGIN
 IF Count>0 THEN
  FOR I:=1 TO Count DO
   P[I]:=C;
 P[0]:=CHR(Count);
END;

PROCEDURE StripChar(VAR S : String;C:CHAR);

VAR I,J,L : LONGINT;

BEGIN
 L:=ORD(S[0]);
 IF L>0 THEN
  BEGIN
   J:=1;
   FOR I:=1 TO L DO
    IF S[I]<>C THEN
     BEGIN
      S[J]:=S[I];
      INC(J);
     END;
   S[0]:=CHR(J-1);
  END;
END;

PROCEDURE KillChrTot(VAR S : String;CONST CSet:CHARSET);

VAR I,J,L : LONGINT;

BEGIN
 L:=ORD(S[0]);
 IF L>0 THEN
  BEGIN
   J:=1;
   FOR I:=1 TO L DO
    IF NOT(S[I] IN CSet) THEN
     BEGIN
      S[J]:=S[I];
      INC(J);
     END;
   S[0]:=CHR(J-1);
  END;
END;

PROCEDURE AppendBackSlash(VAR S :String);

VAR I : LONGINT;

BEGIN
 I:=ORD(S[0]);
{$ifdef linux}
 IF S[I]<>'/' THEN   {Allright for I:=0, If I=0  -> S[0]<>'\'}
  BEGIN
   S[I+1]:='/';
   S[0]:=CHR(I+1);
  END;
{$ELSE}
 IF S[I]<>'\' THEN   {Allright for I:=0, If I=0  -> S[0]<>'\'}
  BEGIN
   S[I+1]:='\';
   S[0]:=CHR(I+1);
  END;
{$ENDIF}
END;

PROCEDURE StripDoubleChar(VAR S : String;C:CHAR);

VAR I,J,L : LONGINT;

BEGIN
 L:=ORD(S[0]);
 IF L>1 THEN  {No double chars to strip when Length(S)=1}
  BEGIN
   J:=2;
   FOR I:=2 TO L DO
    IF NOT((S[I]=C) AND (S[I]=S[I-1])) THEN
     BEGIN
      S[J]:=S[I];
      INC(J);
     END;
   S[0]:=CHR(J-1);
  END;
END;

PROCEDURE UpperCase(VAR S : String);

VAR L,I : LONGINT;

BEGIN
 L:=Length(S);
 IF L>0 THEN
  FOR I:=1 TO L DO
   IF (S[I]>CHR(96)) AND (S[I]<CHR(123)) THEN
    S[I]:=CHR(ORD(S[I])-32);
END;

PROCEDURE LowerCase(VAR S : String);

VAR L,I : LONGINT;

BEGIN
 L:=Length(S);
 IF L>0 THEN
  FOR I:=1 TO L DO
   IF (S[I]>CHR(64)) AND (S[I]<CHR(91)) THEN
    S[I]:=CHR(ORD(S[I])+32);
END;

PROCEDURE Commastr(var S : String;sep:CHAR);

VAR L,I,J : LONGINT;
    SepStr:String[2];

BEGIN
 SepStr[0]:=CHR(1); SepStr[1]:=Sep;
 L:=Length(S);
 I:=L DIV 3;
 IF ((L MOD 3)=0) AND (I>0) THEN
  DEC(I);
 IF I>0 THEN
  FOR J:= 1 TO I DO
    Insert(SepStr,S,L-J*3+1);
END;

PROCEDURE ExpandTabs(CONST P : String;VAR P2:String;Tabsize:LONGINT);

VAR I,B,C,D : LONGINT;

BEGIN
 I:=1; B:=1;
 WHILE I <=ORD(P[0]) DO
  BEGIN
   IF P[I]=CHR(9) THEN
    BEGIN
     C:=Tabsize-((B-1) MOD TabSize);
     FOR D:=0 TO C-1 DO
      P2[B+D]:=' ';
     B:=B+C-1;
    END
   ELSE
    P2[B]:=P[I];
   INC(B); INC(I);
  END;
 P2[0]:=CHR(B);
END;


PROCEDURE BinaryToStr(VAR S : String;Value,Bits : CARDINAL);

VAR I : LONGINT;

BEGIN
 IF Bits >0 THEN
  BEGIN
   IF Bits<>32 THEN
    Value:=Value SHL (32-bits);
   FOR I:=1 TO Bits DO
    BEGIN
     IF (Value AND (1 SHL 31))<>0 THEN
      S[I]:='1'
     ELSE
      S[I]:='0';
    Value:=Value shl 1;
    END;
   END;
  S[0]:=CHR(Bits);
END;

FUNCTION StrToBinary(CONST S : String;Bits : CARDINAL):CARDINAL;

VAR I : CARDINAL;
    A : CARDINAL;

BEGIN
 I:=0;
 IF Bits>0 THEN
  BEGIN
   FOR A :=1 TO Bits DO
    BEGIN
     I:=I SHL 1;
     I:=I OR (ORD(S[A]) AND 1);
    END;
  END;
 StrToBinary:=I;
END;


PROCEDURE HexToStr(VAR S : String;Value,Digits : CARDINAL);

VAR I : LONGINT;
    O : CARDINAL;

BEGIN
 IF Digits >0 THEN
  BEGIN
   IF Digits<>32 THEN
    Value:=Value SHL (32-(Digits shl 2));
   FOR I:=1 TO Digits DO
    BEGIN
     O:=(Value SHR 28) +48;
     IF O>57 THEN
      INC(O,7);
     S[I]:=CHR(O);
    Value:=Value shl 4;
    END;
   END;
  S[0]:=CHR(Digits);
END;


PROCEDURE OctToStr(VAR S : String;Value,Digits : CARDINAL);

VAR I : LONGINT;
    O : CARDINAL;

BEGIN
 IF Digits >0 THEN
  BEGIN
   IF Digits<>32 THEN
    Value:=Value SHL (32-(Digits shl 1 )-Digits);
   FOR I:=1 TO Digits DO
    BEGIN
     O:=(Value SHR 29) +48;
     S[I]:=CHR(O);
    Value:=Value shl 3;
    END;
   END;
  S[0]:=CHR(Digits);
END;


FUNCTION StrToOct(CONST S : String;Digits: CARDINAL):CARDINAL;

VAR I : CARDINAL;
    A : CARDINAL;

BEGIN
 I:=0;
 IF Digits>0 THEN
  BEGIN
   FOR A :=1 TO Digits DO
     I:=(I SHL 3) + ((ORD(S[A])-48) AND 7)
  END;
 StrToOct:=I;
END;

FUNCTION StrToHex(CONST S : String;Digits: CARDINAL):CARDINAL;

VAR I : CARDINAL;
    A : CARDINAL;

BEGIN
 I:=0;
 IF Digits>0 THEN
  BEGIN
   FOR A :=1 TO Digits DO
    BEGIN
     I:=I SHL 4;
     IF S[A]<CHR(58) THEN
      I:=I + ((ORD(S[A])-48) AND 15)
     ELSE
      I:=I + ((ORD(S[A])-55) AND 15);
    END;
  END;
 StrToHex:=I;
END;

PROCEDURE Invert( VAR s :String);

VAR  C          : CHAR;
     Start , En : integer;
BEGIN
        Start:=1;
        En:=Length(s);
        WHILE ( start <= en ) DO
         BEGIN
          c := s[en];
          s[en]  := s[Start];
          s[Start]:=c;
          Start:= Start+1;
          en:= en -1
        END;
END;

{ Same as item. (returns Nth substring of S in R (first is 0), separators are all
characters in S)

However depending on the mode constant this procedure keeps track of quotes.
Mode=0   quote char is "
Mode=1   quote char is '
Mode=2   Modula-2 string mode. both " and ' work, and ' can be inside " and vice versa
         (for strings as the following line (separator =space and comma)
         item1,"item 2 'still item2'"'item3"nonsense" still item 3')

Return= 0  - all ok.
Return=-1 - invalid mode value
Return= other values -  the position of an unmatched quote
}



{$ELSE}
 {$I epas_sa.inc}  {Assembler version ShortString procedures}
{$ENDIF Useasm}

end.
