PROGRAM ReadXF;
{Example program for ConfFiles unit.

   This is a program of the XTDLIB/FPC package
   COPYRIGHT (C) 1993..1998 by Marco van de Voort,  All Rights Reserved
   Free for non-commercial use. See license.txt.
   (www.stack.nl/~marcov/xtdlib.htm)
   Mail for commercial info.

Reads a more complicated file, (/etc/)Xf86config, using the keyword support,
and writes it to a new file. The example seems big, but a quarter is keyword
info. Also remember that the file is analysed in memory between de read and
write.

Problems  :
          - comments before an endsection keyword can move behind it, could
             be solved
          - Keywords get case info of ExampleKeywords "name: field.
          - Distances between keyword and parameter are not the same.
          - Too much code, maybe it would have been better to base it on
               TConfigTree, and using a modified copy of TConfigTreeView.
          - Slow. Finding 10 times the keyword isn't done in one search.

Advantages:
          - Very flexible. A lot of changes/queries can be made with very
             little code. See ReadDevice and ReadMode properties.
          - Still little code for the functionality (complete break-up) it
             does.
}

{$APPTYPE CONSOLE}

Uses EPasStr,ConfFiles;

TYPE KeywordEnum=(Dummy,
    XF_Depth,XF_Modes,XF_Option,XF_BaudRate,XF_BoardName,XF_Buttons,
    XF_Chipset,XF_DefaultColorDepth,XF_Device,XF_Driver,XF_EndSubSection,
    XF_FontPath,XF_HorizSync,XF_Identifier,XF_Modeline,XF_ModelName,
    XF_Monitor,XF_Protocol,XF_Resolution,XF_RgbPath,
    XF_SubSection,XF_VendorName,XF_VertRefresh,XF_VideoRam,XF_XkbLayout,
    XF_XkbModel,XF_XkbOptions,XF_XkbRules,XF_XkbVariant,XF_EndSection,
    XF_Section,XF_AutoRepeat,XF_LeftAlt,XF_RightAlt,XF_ScrollLock,
    XF_RightCtl,XF_XkbKeycodes,XF_XkbTypes,XF_XkbCompat,XF_XkbSymbols,
    XF_XkbGeometry,XF_Emulate3Buttons,XF_Emulate3Timeout,XF_ViewPort);

CONST {Order in this array is NOT important as a result of the CallRoutine field}
ExampleKeywords : ARRAY[XF_Depth..XF_ViewPort] OF KeywRec=(
    (Name:'Depth';               CallRoutine:ORD(XF_Depth);                      Separator : Item_spacecomma;   Functns:[]),
    (Name:'Modes';               CallRoutine:ORD(XF_Modes);                      Separator : Item_spacecomma;   Functns:[]),
    (Name:'Option';              CallRoutine:ORD(XF_Option);                     Separator : Item_spacecomma;   Functns:[]),
    (Name:'BaudRate';            CallRoutine:ORD(XF_BaudRate);                   Separator : Item_spacecomma;   Functns:[]),
    (Name:'BoardName';           CallRoutine:ORD(XF_BoardName);                  Separator : Item_spacecomma;   Functns:[]),
    (Name:'Buttons';             CallRoutine:ORD(XF_Buttons);                    Separator : Item_spacecomma;   Functns:[]),
    (Name:'Chipset';             CallRoutine:ORD(XF_Chipset);                    Separator : Item_spacecomma;   Functns:[]),
    (Name:'DefaultColorDepth';   CallRoutine:ORD(XF_DefaultColorDepth);          Separator : Item_spacecomma;   Functns:[]),
    (Name:'Device';              CallRoutine:ORD(XF_Device);                     Separator : Item_spacecomma;   Functns:[]),
    (Name:'Driver';              CallRoutine:ORD(XF_Driver);                     Separator : Item_spacecomma;   Functns:[]),
    (Name:'EndSubSection';       CallRoutine:ORD(XF_EndSubSection);              Separator : Item_spacecomma;   Functns:[AllowZero]),
    (Name:'FontPath';            CallRoutine:ORD(XF_FontPath);                   Separator : Item_spacecomma;   Functns:[]),
    (Name:'HorizSync';           CallRoutine:ORD(XF_HorizSync);                  Separator : Item_spacecomma;   Functns:[]),
    (Name:'Identifier';          CallRoutine:ORD(XF_Identifier);                 Separator : Item_spacecomma;   Functns:[]),
    (Name:'Modeline';            CallRoutine:ORD(XF_Modeline);                   Separator : Item_spacecomma;   Functns:[]),
    (Name:'ModelName';           CallRoutine:ORD(XF_ModelName);                  Separator : Item_spacecomma;   Functns:[]),
    (Name:'Monitor';             CallRoutine:ORD(XF_Monitor);                    Separator : Item_spacecomma;   Functns:[]),
    (Name:'Protocol';            CallRoutine:ORD(XF_Protocol);                   Separator : Item_spacecomma;   Functns:[]),
    (Name:'Resolution';          CallRoutine:ORD(XF_Resolution);                 Separator : Item_spacecomma;   Functns:[]),
    (Name:'RgbPath';             CallRoutine:ORD(XF_RgbPath);                    Separator : Item_spacecomma;   Functns:[]),
    (Name:'SubSection';          CallRoutine:ORD(XF_SubSection);                 Separator : Item_spacecomma;   Functns:[]),
    (Name:'VendorName';          CallRoutine:ORD(XF_VendorName);                 Separator : Item_spacecomma;   Functns:[]),
    (Name:'VertRefresh';         CallRoutine:ORD(XF_VertRefresh);                Separator : Item_spacecomma;   Functns:[]),
    (Name:'VideoRam';            CallRoutine:ORD(XF_VideoRam);                   Separator : Item_spacecomma;   Functns:[]),
    (Name:'XkbLayout';           CallRoutine:ORD(XF_XkbLayout);                  Separator : Item_spacecomma;   Functns:[]),
    (Name:'XkbModel';            CallRoutine:ORD(XF_XkbModel);                   Separator : Item_spacecomma;   Functns:[]),
    (Name:'XkbOptions';          CallRoutine:ORD(XF_XkbOptions);                 Separator : Item_spacecomma;   Functns:[]),
    (Name:'XkbRules';            CallRoutine:ORD(XF_XkbRules);                   Separator : Item_spacecomma;   Functns:[]),
    (Name:'XkbVariant';          CallRoutine:ORD(XF_XkbVariant);                 Separator : Item_spacecomma;   Functns:[]),
    (Name:'EndSection';          CallRoutine:ORD(XF_EndSection);                 Separator : Item_spacecomma;   Functns:[AllowZero]),
    (Name:'Section';             CallRoutine:ORD(XF_Section);                    Separator : Item_spacecomma;   Functns:[]),
    (Name:'AutoRepeat';          CallRoutine:ORD(XF_AutoRepeat);                 Separator : Item_spacecomma;   Functns:[]),
    (Name:'LeftAlt';             CallRoutine:ORD(XF_LeftAlt);                    Separator : Item_spacecomma;   Functns:[]),
    (Name:'RightAlt';            CallRoutine:ORD(XF_RightAlt);                   Separator : Item_spacecomma;   Functns:[]),
    (Name:'ScrollLock';          CallRoutine:ORD(XF_ScrollLock);                 Separator : Item_spacecomma;   Functns:[]),
    (Name:'RightCtl';            CallRoutine:ORD(XF_RightCtl);                   Separator : Item_spacecomma;   Functns:[]),
    (Name:'XkbKeycodes';         CallRoutine:ORD(XF_XkbKeycodes);                Separator : Item_spacecomma;   Functns:[]),
    (Name:'XkbTypes';            CallRoutine:ORD(XF_XkbTypes);                   Separator : Item_spacecomma;   Functns:[]),
    (Name:'XkbCompat';           CallRoutine:ORD(XF_XkbCompat);                  Separator : Item_spacecomma;   Functns:[]),
    (Name:'XkbSymbols';          CallRoutine:ORD(XF_XkbSymbols);                 Separator : Item_spacecomma;   Functns:[]),
    (Name:'XkbGeometry';         CallRoutine:ORD(XF_XkbGeometry);                Separator : Item_spacecomma;   Functns:[]),
    (Name:'Emulate3Buttons';     CallRoutine:ORD(XF_Emulate3Buttons);            Separator : Item_spacecomma;   Functns:[]),
    (Name:'Emulate3Timeout';     CallRoutine:ORD(XF_Emulate3Timeout);            Separator : Item_spacecomma;   Functns:[]),
    (Name:'ViewPort';            CallRoutine:ORD(XF_ViewPort);                   Separator : Item_spacecomma;   Functns:[])    );

TYPE TXF86Config=CLASS(TConfigTreeView)
                    private
                     InSection: BOOLEAN;
                     SectionOpen : SET OF 1..20;
                     PROCEDURE FinalisationHelper(Level:LONGINT);
                    protected
                     FUNCTION ReadMode(Ind:LONGINT): String;
                     FUNCTION ReadDevice(Ind:LONGINT): String;
                    public
                     CONSTRUCTOR NewConfigFile; OVERRIDE;
                     FUNCTION  DoKeyWord(KeywCode:LONGINT;Params:ParamType;ItemNr:LONGINT): LONGINT; override;
                     PROCEDURE WriteValue(Handle:PLineData;Level:LONGINT);override;
                     PROCEDURE BeforeExitFile; Override;
                     PROCEDURE WriteSection(Section : PSection); override;
                     property GetMode[Ind:LONGINT]:String read readmode;
                     property GetDevice[Ind:LONGINT]:String read readDevice;
                    END;

CONSTRUCTOR TXF86Config.NewConfigFile;

BEGIN
 Inherited NewConfigFile;
 InSection:=FALSE;                     {Writes before first section give error}
END;

FUNCTION TXF86Config.ReadMode(Ind:LONGINT): String;

VAR Sect : PSection;
    S,S2 : String;

BEGIN
 Sect:=GlobalSection['DISPLAY',Ind];
 IF (Sect<>NIL) AND (Sect^.LineData<>NIL) THEN
  BEGIN
   S:=LocalKeyword[ORD(XF_Modes),Sect,1];
   S2:=LocalKeyword[ORD(XF_Depth),Sect,1];
   LGrow(S2,' ',2);
   S:=S+','+S2;
   Sect:=Sect^.ParentSection;
   S:=S+','+LocalKeyword[ORD(XF_Driver),Sect,1];
   ReadMode:=S;
  END
 ELSE
  ReadMode:=ValueNotAvailable;
END;

FUNCTION TXF86Config.ReadDevice(Ind:LONGINT): String;

VAR Sect : PSection;
    S    : String;

BEGIN
 Sect:=GlobalSection['Device',Ind];
 IF (Sect<>NIL) AND (Sect^.LineData<>NIL) THEN
  BEGIN
   S:=LocalKeyword[ORD(XF_Identifier),Sect,1];
   RGrow(S,' ',17);
   S:=S+LocalKeyword[ORD(XF_VendorName),Sect,1];
   RGrow(S,' ',34);
   S:=S+LocalKeyword[ORD(XF_BoardName),Sect,1];
   RGrow(S,' ',51);
   S:=S+LocalKeyword[ORD(XF_Chipset),Sect,1];
   RGrow(S,' ',68);
   S:=S+LocalKeyword[ORD(XF_VideoRam),Sect,1];
   ReadDevice:=S;
  END
 ELSE
  ReadDevice:=ValueNotAvailable;

END;

PROCEDURE TXF86Config.WriteValue(Handle:PLineData;Level:LONGINT);

VAR S : String;
    I,A : LONGINT;
    P : PParamData;
    PD : ParamType;

PROCEDURE RollOutParam;

BEGIN
   P:=Handle^.ParamData;
   WHILE P<>NIL DO
    BEGIN
     PD:=GetNextParam(P);
     IF NOT((PD.ValueType=0) AND (PD.Value='')) THEN
        CASE PD.ValueType OF
         0 : Write(FText,PD.Value);
         1 : Write(FText,PD.Value);
        END;
     IF P<>NIL THEN
      Write(FText,' ');
    END;
END;

BEGIN
 IF Handle^.KeyNo<>-1 THEN
  BEGIN
    i:=CallRoutineToIndex(Handle^.KeyNo);
    Write(FText,' ':(Level*3));
    Write(FText,ExampleKeywords[KeywordEnum(I)].Name, ' ');
    I:=KeywordIndent-Length(keyw[I].Name);
    IF (I>0) THEN
    Write(FText,' ':I);
    RollOutParam;
  END
 ELSE
  BEGIN
   Write(FText,Handle^.ValueName);
   I:=IniIndent-Length(Handle^.ValueName);
   IF (I>0)  THEN
    BEGIN
     Write(FText,' ':I);
     I:=0;
    END;
   IF IsIni THEN
    BEGIN
     Write (FText,KeySepPref);
     I:=I+ParamIndent;
     IF (i>0) THEN
      Write(FText,' ':I);
     END;
    RollOutParam;
  END;
 Writeln(FText);
END;

PROCEDURE TXF86Config.FinalisationHelper(Level:LONGINT);

VAR A : LONGINT;

BEGIN
   A:=20;
   WHILE (A>=Level) AND (A>0) DO
    BEGIN
     IF A IN SectionOpen THEN
      BEGIN
      IF A=1 THEN
       WriteLn(FText,'EndSection')
      ELSE
       BEGIN
        Write(FText,' ':(A-1)*3);
        WriteLn(FText,'EndSubSection')
       END;
      EXCLUDE(SectionOpen,A);
     END;
     DEC(A);
    END;
END;

PROCEDURE TXF86Config.BeforeExitFile;

BEGIN
 FinalisationHelper(1);
END;


PROCEDURE TXF86Config.WriteSection(Section : PSection);

VAR A :LONGINT;

BEGIN
 IF Section^.SectionName<>'ROOTSECTION' THEN
  BEGIN
   FinalisationHelper(Section^.Level);
   WriteComment(Section^.Comments);
   INCLUDE(SectionOpen,SectioN^.Level);
   IF Section^.Level=1 THEN
    Write(FText,'Section ')
   ELSE
    BEGIN
     Write(FText,' ':(Section^.Level-1)*3);
     Write(FText,'SubSection ')
    END;
   WriteLn(FText,Section^.SectionName);
  END;
END;


FUNCTION  TXF86Config.DoKeyWord(KeywCode:LONGINT;Params:ParamType;ItemNr:LONGINT): LONGINT;

VAR I : LONGINT;
    S : String;

BEGIN
 I:=CallRoutineToIndex(Keywcode);
 LTrim(Params.Value,'"');
 RTrim(Params.Value,'"');
 IF KeywCode=-1 THEN {Store unknown keywords as INI values without further preprocessing}
  I:=Inherited DoKeyword(Keywcode,params,0)
 ELSE
  BEGIN
 CASE  KeywordEnum(KeywCode) OF
  XF_EndSubSection: BEGIN
                     IF CurSection^.Level<2 THEN
                      BEGIN
                       Writeln('No subsection to end');
                       Writeln('Line:',Line);
                       HALT;
                      END;
                     DoSection('',CurSection^.Level-1); {go to lower section}
                    END;
  XF_EndSection:   BEGIN
                     IF CurSection^.Level<>1 THEN
                      BEGIN
                       Writeln('Error can''t go below this level line:',Line);
                       HALT;
                      END;
                     InSection:=FALSE;  {Don't allow keywords until next Section}
                    END;
   XF_SubSection : BEGIN
                     IF NOT InSection THEN
                      BEGIN
                       Writeln('We are not inside an section command.');
                       HALT;
                      END;
                    DoSection(UPCASE(Params.Value),CurSection^.Level+1);
                   END;
   XF_Section:     BEGIN
                    IF InSection THEN
                      BEGIN
                       DoSection(UPCASE(Params.Value),CurSection^.Level+1);
                       WritelN('Warning, section used to nest in line: ',Line);
                      END
                      ELSE
                       BEGIN
                        InSection:=TRUE;
                        DoSection(UPCASE(Params.Value),1);
                     END;
                    END
        ELSE
         BEGIN       
          inherited Dokeyword(keywcode,Params,0);
         END;
        END;
     END;
END;

Var XFCfg : TXF86Config;
    S            : String;
    A            : LONGINT;
    Sect         : PSection;

BEGIN
 XFCfg:=TXF86Config.NewConfigFile;
 XFCfg.SetKeywords(ExampleKeywords);
 XFCfg.IsIni:=FALSE;
 XFCfg.HoldSectionComment:=TRUE;
 INCLUDE(XFCfg.Commentset,'#');
 XFCfg.ReadConfig('Xf86Config');
 Writeln('Videocards:');
 Writeln;

 WriteLn('Identifier       VendorName       BoardName        Chipset          VideoRam');
 A:=1;
 REPEAT
  S:=XFCfg.GetDevice[A];
  IF S<>ValueNotAvailable THEN
    Writeln(S);
  INC(A);
 UNTIL S=ValueNotAvailable;
 Writeln;
 Writeln('Modi:');
 Writeln;
 A:=1;
 REPEAT
  S:=XFCfg.GetMode[A];
  IF S<>ValueNotAvailable THEN
   Writeln(S);
  INC(A);
 UNTIL S=ValueNotAvailable;

 A:=1;
 REPEAT
  S:=XFCfg.GlobalGenericKey[A];
  IF S<>ValueNotAvailable THEN
   BEGIN
    IF A=1 THEN
     BEGIN
      Writeln;
      Writeln('Keywords not recognized by this program:');
      Writeln;
     END;
    Writeln('Unknown keyword: ', S);
   END;
  INC(A);
 UNTIL S=ValueNotAvailable;
 XFCfg.WriteTree('newxf86config');
 XFCfg.KillTree;
 XFCfg.KillConfigFile;
END.
